# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Implements a single API for instrumenting all dbapi2-compliant modules
"""
from contrast.agent.policy import patch_manager
from contrast.applies.sqli import apply_rule
from contrast.utils.decorators import fail_safely
from contrast.utils.patch_utils import patch_cls_or_instance


def _build_patch(database, action):
    def patched_method(orig_func, patch_policy, *args, **kwargs):
        return apply_rule(database, action, orig_func, args, kwargs)

    return patched_method


def _instrument_cursor_method(database, cursor, method_name):
    new_method = _build_patch(database, method_name)
    patch_cls_or_instance(cursor, method_name, new_method)
    func = patch_manager.as_func(getattr(cursor, method_name))
    func.__name__ = method_name


def instrument_cursor(database, cursor):
    """
    Instruments a dbapi2-compliant database cursor class

    @param database: Name of the database module being patched (e.g. "sqlite3")
    @param cursor: Reference to cursor class to be instrumented
    """
    _instrument_cursor_method(database, cursor, "execute")
    _instrument_cursor_method(database, cursor, "executemany")


def instrument_executescript(database, cursor):
    """
    Instruments the `executescript` method of a database cursor class

    The executescript method is non-standard but is provided by some drivers
    including sqlite3.

    @param database: Name of the database module being patched (e.g. "sqlite3")
    @param cursor: Reference to cursor class to be instrumented
    """
    _instrument_cursor_method(database, cursor, "executescript")


@fail_safely("failed to instrument database adapter")
def instrument_adapter(database, adapter):
    """
    In some cases (SQLAlchemy), we need to instrument an unknown PEP-249 compliant
    adapter. We only have a reference to the adapter module, and we can't make any
    assumptions about the existence of `adapter.Cursor`, since this is not guaranteed
    by the spec.

    We are only guaranteed the following:
    - the adapter has a `connection()` method, which returns an instance of Connection
    - the Connection object has a `cursor()` method, which returns an instance of Cursor
    - the Cursor has `execute()` and `executemany()` methods

    This requires a somewhat roundabout instrumentation strategy:
    - on the first call to adapter.connect(), we can access the Connection class
    - on the first call to Connection.cursor(), we can access the Cursor class
    - this lets us instrument Cursor.execute() and Cursor.executemany()
    """

    @fail_safely("failed to instrument database cursor object")
    def safe_instrument_cursor(cursor_instance):
        """
        Safely instrument a Cursor class given an instance of that class
        """
        cursor_class = type(cursor_instance)
        if not patch_manager.is_patched(cursor_class.execute):
            instrument_cursor(database, cursor_class)

    @fail_safely("failed to instrument database connection object")
    def safe_instrument_connection(connection_instance):
        """
        Safely instrument a Connection class given an instance of that class
        """
        connection_class = type(connection_instance)
        if not patch_manager.is_patched(connection_class.cursor):
            patch_cls_or_instance(connection_class, "cursor", cursor_patch)

    def cursor_patch(orig_func, _, *args, **kwargs):
        """
        Patch for dbapi_adapter.connection().cursor()

        This patch will ensure that the returned Cursor object's class will have
        `execute` and `executemany` instrumented.
        """
        cursor = orig_func(*args, **kwargs)
        safe_instrument_cursor(cursor)
        return cursor

    def connect_patch(orig_func, _, *args, **kwargs):
        """
        Patch for dbapi_adapter.connection()

        This patch will ensure that the returned Connection object's class will have
        `cursor_patch` applied to its cursor() method.
        """
        connection = orig_func(*args, **kwargs)
        safe_instrument_connection(connection)
        return connection

    if not patch_manager.is_patched(adapter.connect):
        patch_cls_or_instance(adapter, "connect", connect_patch)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast.extern.wrapt import CallableObjectProxy, register_post_import_hook

from contrast.agent.assess.policy.preshift import Preshift
from contrast.agent.assess.policy.analysis import _analyze
from contrast.agent.policy import patch_manager
from contrast.agent.policy.loader import Policy
from contrast.utils.decorators import fail_safely
from contrast.utils.patch_utils import patch_cls_or_instance


@fail_safely("Failed to apply assess xpath-injection")
def apply_assess(location, self, retval, args, kwargs):
    policy = Policy()
    patch_policy = policy.policy_by_name.get(location)
    if patch_policy is None:
        return

    preshift = Preshift(self, args, kwargs)
    _analyze(patch_policy, preshift, self, retval, (self,) + args, kwargs)


def apply_call(class_name, orig_func, self, args, kwargs):
    result = None
    try:
        result = orig_func(*args, **kwargs)
    finally:
        location = "lxml.etree.{}.__call__".format(class_name)
        apply_assess(location, self, result, args, kwargs)
    return result


class ContrastXPathEvaluatorProxy(CallableObjectProxy):
    """
    Proxy class that wraps instances returned by XPathEvaluator factory

    We instrument the relevant classes directly, but since the factory is implemented
    as a C extension, the instances that it returns are the original type instead of
    our replacement. In order to cover all of our bases, we need both the replacement
    subclass and a proxied class that we return from the instrumented factory.
    """

    def __call__(self, *args, **kwargs):
        self_obj = self.__wrapped__
        orig_func = self.__wrapped__.__call__
        return apply_call(
            self_obj.__class__.__name__, orig_func, self_obj, args, kwargs
        )


def create_instrumented_xpath_element_evaluator(XPathElementEvaluator):
    """
    Generate instrumented subclass of XPathElementEvaluator

    We can't simply declare this at module level since we can't guarantee that lxml
    will be installed. We need to wait until the import hook is fired to know whether
    it's safe to make a reference to the original type.
    """

    class ContrastXPathElementEvaluator(XPathElementEvaluator):
        def __call__(self, *args, **kwargs):
            orig_func = super(ContrastXPathElementEvaluator, self).__call__
            return apply_call(
                XPathElementEvaluator.__name__, orig_func, self, args, kwargs
            )

    return ContrastXPathElementEvaluator


def create_instrumented_xpath_document_evaluator(XPathDocumentEvaluator):
    """
    Generate instrumented subclass of XPathDocumentEvaluator

    See docstring for create_instrumented_xpath_element_evaluator above.
    """

    class ContrastXPathDocumentEvaluator(XPathDocumentEvaluator):
        def __call__(self, *args, **kwargs):
            orig_func = super(ContrastXPathDocumentEvaluator, self).__call__
            return apply_call(
                XPathDocumentEvaluator.__name__, orig_func, self, args, kwargs
            )

    return ContrastXPathDocumentEvaluator


def create_instrumented_xpath(XPath):
    """
    Generate instrumented subclass of XPath

    We can't simply declare this at module level since we can't guarantee that lxml
    will be installed. We need to wait until the import hook is fired to know whether
    it's safe to make a reference to the original type.
    """

    class ContrastXPath(XPath):
        def __init__(self, *args, **kwargs):
            try:
                super(ContrastXPath, self).__init__(*args, **kwargs)
            finally:
                apply_assess("lxml.etree.XPath.__init__", self, None, args, kwargs)

    return ContrastXPath


def XPathEvaluator(orig_func, patch_policy, *args, **kwargs):
    """
    Instrumented version of XPathEvaluator factory
    """
    evaluator = orig_func(*args, **kwargs)
    return ContrastXPathEvaluatorProxy(evaluator)


def patch_etree(etree_module):
    patch_cls_or_instance(etree_module, "XPathEvaluator", XPathEvaluator)

    new_xpath = create_instrumented_xpath(etree_module.XPath)
    patch_manager.patch(etree_module, "XPath", new_xpath)

    new_element_evaluator = create_instrumented_xpath_element_evaluator(
        etree_module.XPathElementEvaluator
    )
    patch_manager.patch(etree_module, "XPathElementEvaluator", new_element_evaluator)

    new_document_evaluator = create_instrumented_xpath_document_evaluator(
        etree_module.XPathDocumentEvaluator
    )
    patch_manager.patch(etree_module, "XPathDocumentEvaluator", new_document_evaluator)


def register_patches():
    register_post_import_hook(patch_etree, "lxml.etree")


def reverse_patches():
    module = sys.modules.get("lxml.etree")
    if not module:
        return

    patch_manager.reverse_patches_by_owner(module)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import mock
import threading

import contrast

from contrast.api.settings_pb2 import AgentSettings, Reaction
from contrast.agent.settings_state import SettingsState


def assert_empty_context_tracker():
    context_tracker = contrast.CS__CONTEXT_TRACKER

    assert not any(
        context_tracker.get(thread_id, None) for thread_id in context_tracker._tracker
    )


def build_agent_settings_shutdown():
    agent_settings = AgentSettings()
    reaction = Reaction()

    reaction.operation = Reaction.DISABLE
    agent_settings.application_settings.reactions.append(reaction)

    return agent_settings


def thread_running_by_id(thread_id):
    return any(id(thread) == thread_id for thread in threading.enumerate())


class AgentShutdownMixin:
    @mock.patch(
        "contrast.agent.middlewares.base_middleware.BaseMiddleware.call_without_agent"
    )
    def test_agent_shutdown_in_heartbeat_thread(self, mock_call_without_agent, mocker):
        settings = SettingsState()
        trigger = "os-system"
        user_input = "whatever"
        path = "/vulnpy/cmdi/{}".format(trigger)

        response = self.client.get(path, {"user_input": user_input})

        assert response.status_code == 200
        assert settings.is_agent_config_enabled()
        assert thread_running_by_id(id(settings.heartbeat))
        assert not mock_call_without_agent.called

        mocked_send_dtm = mocker.patch(
            "contrast.agent.service_client._ServiceState.send_dtm",
            return_value=build_agent_settings_shutdown(),
        )

        settings.heartbeat.join(timeout=10)

        assert mocked_send_dtm.called
        assert not settings.is_agent_config_enabled()
        assert not thread_running_by_id(id(settings.heartbeat))

        assert_empty_context_tracker()

    @mock.patch(
        "contrast.agent.middlewares.base_middleware.BaseMiddleware.call_without_agent"
    )
    def test_agent_shutdown_in_request_thread(self, mock_call_without_agent, mocker):
        settings = SettingsState()
        trigger = "os-system"
        user_input = "whatever"
        path = "/vulnpy/cmdi/{}".format(trigger)

        response = self.client.get(path, {"user_input": user_input})

        assert response.status_code == 200
        assert settings.is_agent_config_enabled()
        assert thread_running_by_id(id(settings.heartbeat))
        assert not mock_call_without_agent.called

        mocker.patch("contrast.agent.heartbeat.Heartbeat.send_heartbeat")
        mocked_send_dtm = mocker.patch(
            "contrast.agent.service_client._ServiceState.send_dtm",
            return_value=build_agent_settings_shutdown(),
        )

        response = self.client.get(path, {"user_input": user_input})

        assert response.status_code == 200
        assert mocked_send_dtm.called
        assert not settings.is_agent_config_enabled()

        mock_track_str = mocker.patch("contrast.agent.assess.string_tracker")

        response = self.client.get(path, {"user_input": user_input})

        assert response.status_code == 200
        assert not mock_track_str.called
        assert mocked_send_dtm.called
        assert mock_call_without_agent.called
        assert not settings.is_agent_config_enabled()

        settings.heartbeat.join(timeout=10)

        assert not thread_running_by_id(id(settings.heartbeat))

        assert_empty_context_tracker()



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os

from contrast.agent.framework import Framework
from contrast.agent.heartbeat import Heartbeat
from contrast.agent.protect.mixins.REP_settings import SettingsREPMixin
from contrast.agent.protect.rule.rules_builder import RulesBuilder
from contrast.agent.reaction_processor import ReactionProcessor
from contrast.api.settings_pb2 import (
    AccumulatorSettings,
    AgentSettings,
    ApplicationSettings,
    ServerFeatures,
)
from contrast.configuration import AgentConfig
from contrast.utils.loggers.logger import reset_agent_logger
from contrast.utils.singleton import Singleton
from contrast.utils.string_utils import truncate

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class SettingsState(Singleton, SettingsREPMixin):
    def init(self, app_name=None):
        """
        Singletons should override init, not __init__.
        """
        self.config = None
        self.config_features = {}
        self.last_update = None
        self.heartbeat = None
        self.framework = Framework()
        self.sys_module_count = 0

        # Features and Settings from Service
        self.server_features = ServerFeatures()
        self.application_settings = ApplicationSettings()
        self.accumulator_settings = AccumulatorSettings()

        # Server
        self.server_name = None
        self.server_path = None
        self.server_type = None

        self.exclusion_matchers = []

        # Rules
        self.defend_rules = dict()

        # Initialize config
        self.config = AgentConfig()

        self.disabled_assess_rules = set(
            self.config.get("assess.rules.disabled_rules", [])
        )

        # Initialize application metadata
        self.app_name = self.get_app_name(app_name)

        logger.info("Contrast Agent finished loading settings.")

    def _is_defend_enabled_in_server_features(self):
        return (
            self.server_features
            and self.server_features.defend
            and self.server_features.defend.enabled
        )

    def is_agent_config_enabled(self):
        return self.config.get("enable", True)

    @property
    def is_bundled_enabled(self):
        """The default across agents is to use the bundled service by default"""
        return self.config.get("agent.service.enable", True)

    @property
    def pid(self):
        """
        pid is used by Speedracer to recognize a unique worker process for an application.

        pid must be unique for each worker process of an app.
        :return: int current process id
        """
        return os.getpid()

    def get_app_name(self, app_name):
        # Used in DARPA-Screener to set the app's name (so TS can tell the difference between 5 instances of the app)
        if os.environ.get("CONTRASTSECURITY_APP_NAME"):
            return os.environ.get("CONTRASTSECURITY_APP_NAME")

        if self.config.get("application.name"):
            return self.config.get("application.name")

        return app_name if app_name else "root"

    def establish_heartbeat(self):
        """
        Initialize Heartbeat between Agent and SR if it has not been already initialized.
        """
        if self.heartbeat is None:
            self.heartbeat = Heartbeat(self)
            self.heartbeat.start()

    def process_responses(self, responses):
        """
        :param responses: list of Message responses from SR
        """
        self.establish_heartbeat()

        logger.debug("Processing %s responses", len(responses))

        for response in responses:
            if self.process_service_response(response):
                self.set_protect_rules()

    def process_service_response(self, data):
        reload_rules = False

        if data and isinstance(data, AgentSettings):
            self.last_update = data.sent_ms

            reload_rules = self.process_server_features(data) or reload_rules
            reload_rules = self.process_application_settings(data) or reload_rules
            reload_rules = self.process_accumulator_settings(data) or reload_rules

        if reload_rules:
            logger.debug(
                "Finished processing Contrast Service message and reloading rules."
            )

        return reload_rules

    def process_server_features(self, data):
        if not data.HasField("server_features"):
            return False

        self.server_features = data.server_features
        self.update_logger_from_features()

        self.log_server_features(data.server_features)

        return True

    def log_server_features(self, server_features):
        """
        Record server features received from teamserver (via the contrast service)
        Remove the rule_definitions field before logging because it's long and ugly
        """
        server_features_copy = ServerFeatures()
        server_features_copy.CopyFrom(server_features)
        del server_features_copy.defend.rule_definitions[:]
        logger.debug(
            "Received updated server features (excluding rule_definitions from"
            " log):\n%s",
            server_features_copy,
        )

    @property
    def code_exclusion_matchers(self):
        return [x for x in self.exclusion_matchers if x.is_code]

    def process_application_settings(self, data):
        if not data.HasField("application_settings"):
            return False

        self.application_settings = data.application_settings

        ReactionProcessor.process(data.application_settings, self)

        self.reset_transformed_settings()

        logger.debug(
            "Received updated application settings:\n%s", data.application_settings
        )

        return True

    def reset_transformed_settings(self):
        self.exclusion_matchers = []

    def process_accumulator_settings(self, data):
        if data.HasField("accumulator_settings"):
            self.accumulator_settings = data.accumulator_settings

    def update_logger_from_features(self):
        # Python does not support TRACE level so we use DEBUG for now
        log_level = (
            "DEBUG"
            if self.server_features.log_level == "TRACE"
            else self.server_features.log_level
        )

        logger_reset = reset_agent_logger(self.server_features.log_file, log_level)

        if logger_reset:
            self.config.log_config()

    def is_inventory_enabled(self):
        """
        inventory.enable = false: Disables both route coverage and library analysis and reporting
        """
        return self.config.get("inventory.enable", True)

    def is_analyze_libs_enabled(self):
        """
        inventory.analyze_libraries = false: Disables only library analysis/reporting
        """
        return (
            self.config.get("inventory.analyze_libraries", True)
            and self.is_inventory_enabled()
        )

    def is_assess_enabled(self):
        """
        We do not allow assess and defend to be on at the same time. As defend
        is arguably the more important of the two, it will take precedence

        The agent config may enable assess even if it is turned off in TS. This
        allows unlicensed apps to send findings to TS, where they will appear
        as obfuscated results.
        """
        if self.config is None or self.is_protect_enabled():
            return False

        assess_enabled = self.config.get("assess.enable", None)
        if assess_enabled is not None:
            return assess_enabled

        return (
            self.server_features
            and self.server_features.assess
            and self.server_features.assess.enabled
        )

    def is_protect_enabled(self):
        """
        Protect is enabled only if both configuration and server features enable it.
        """
        if self.config is None:
            return False
        config_protect_enabled = self.config.get("protect.enable", True)
        return config_protect_enabled and self._is_defend_enabled_in_server_features()

    def set_protect_rules(self):
        """
        Stores all of our defend rules

        """
        if not self._is_defend_enabled_in_server_features():
            self.defend_rules = dict()
            return
        self.defend_rules = RulesBuilder.build_protect_rules(self)
        logger.debug("Built %s protect rules from settings", len(self.defend_rules))

    def get_server_name(self):
        """
        Hostname of the server

        Default is socket.gethostname() or localhost
        """
        if self.server_name is None:
            self.server_name = self.config.get("server.name")

        return self.server_name

    def get_server_path(self):
        """
        Working Directory of the server

        Default is root
        """
        if self.server_path is None:
            self.server_path = self.config.get("server.path") or truncate(os.getcwd())

        return self.server_path

    def get_server_type(self):
        """
        Web Framework of the Application either defined in common config or via discovery.
        """
        if self.server_type is None:
            self.server_type = (
                self.config.get("server.type") or self.framework.name_lower
            )

        return self.server_type

    @property
    def response_scanning_enabled(self):
        return self.is_assess_enabled() and self.config.get(
            "assess.enable_scan_response"
        )

    def is_assess_rule_disabled(self, rule_id):
        """
        Rules disabled in config override all disabled rules from TS per common config
        """
        return (
            rule_id in self.disabled_assess_rules
            if self.disabled_assess_rules
            else rule_id in self.application_settings.disabled_assess_rules
        )



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern.wrapt import register_post_import_hook

from contrast.patches.databases import dbapi2
from contrast.patches.databases.pymysql_patch import patch_connect


MYSQL_CONNECTOR = "mysql.connector"
VENDOR = "MySQL"


def instrument_mysql_connector(mysql_connector):
    """
    Add patches for mysql.connector

    mysql.connector.cursor.MySQLCursor is the type returned by connect. However, we
    also patch mysql.connector.curstor.CursorBase just to cover our bases (ha!). This
    could potentially be useful if there are other cursor types that inherit
    CursorBase. However, patching that alone does not appear to be sufficient for
    instrumenting MySQLCursor, which is why we have both.
    """
    patch_connect(mysql_connector)

    dbapi2.instrument_cursor(MYSQL_CONNECTOR, mysql_connector.cursor.CursorBase)
    dbapi2.instrument_cursor(MYSQL_CONNECTOR, mysql_connector.cursor.MySQLCursor)


def register_patches():
    register_post_import_hook(instrument_mysql_connector, MYSQL_CONNECTOR)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

# TODO: PYT-1213 this import will eventually be deprecated
from contrast.falcon.middleware import FalconMiddleware  # pylint: disable=unused-import



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os

from contrast.extern.six.moves.urllib_parse import unquote

from contrast.api.dtm_pb2 import PathTraversalSemanticAnalysisDetails
from contrast.utils.decorators import cached_property
from contrast.utils.stack_trace_utils import StackTraceUtils

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class PathTraversalSinkFeatures(object):
    KNOWN_SECURITY_BYPASS_MARKERS = ["::$DATA", "::$Index", "" + "\x00"]

    CUSTOM_CODE_CONFIG_KEY = "detect_custom_code_accessing_system_files"
    COMMON_FILE_EXPLOITS_KEY = "detect_common_file_exploits"
    SYSTEM_FILES = [
        "/proc/self",
        "etc/passwd",
        "etc/shadow",
        "etc/hosts",
        "etc/groups",
        "etc/gshadow",
        "ntuser.dat",
        "/Windows/win.ini",
        "/windows/system32/",
        "/windows/repair/",
    ]

    def check_sink_features(self, attack_vector, attack):
        new_sample = self._check_custom_code_accessing_system_files(
            attack_vector
        ) or self._check_file_security_bypass(attack_vector)

        if new_sample:
            if attack is None:
                attack = self.build_base_attack()

            attack.samples.extend([new_sample])
            attack.response = self.response_from_mode(self.mode)
            self.log_rule_matched(None, attack.response)

        return attack

    def _check_custom_code_accessing_system_files(self, attack_vector):
        if (
            self._is_custom_code_access_sysfile_enabled
            and PathTraversalSinkFeatures._is_custom_code_accessing_file_system(
                attack_vector
            )
        ):
            logger.debug(
                "Found custom code trying to access system file: %s", attack_vector
            )
            return self._create_path_traversal_sample(
                attack_vector,
                PathTraversalSemanticAnalysisDetails.CUSTOM_CODE_ACCESSING_SYSTEM_FILES,
            )

        return None

    def _check_file_security_bypass(self, attack_vector):
        if (
            self._is_common_file_exploits_enabled
            and PathTraversalSinkFeatures._contains_known_attack_signatures(
                attack_vector
            )
        ):
            logger.debug(
                "Found attempt to bypass file security checks: %s", attack_vector
            )
            return self._create_path_traversal_sample(
                attack_vector,
                PathTraversalSemanticAnalysisDetails.COMMON_FILE_EXPLOITS,
            )

        return None

    def _create_path_traversal_sample(self, attack_vector, rep_code):
        sample = self.build_sample(None, attack_vector)
        sample.path_traversal_semantic.path = attack_vector
        sample.path_traversal_semantic.findings.append(rep_code)
        return sample

    @cached_property
    def _is_custom_code_access_sysfile_enabled(self):
        return self.settings.is_rep_feature_enabled_for_rule(
            self.name, self.CUSTOM_CODE_CONFIG_KEY
        )

    @cached_property
    def _is_common_file_exploits_enabled(self):
        return self.settings.is_rep_feature_enabled_for_rule(
            self.name, self.COMMON_FILE_EXPLOITS_KEY
        )

    @staticmethod
    def _is_custom_code_accessing_file_system(path):
        return (
            PathTraversalSinkFeatures._is_system_file(path)
            and StackTraceUtils.in_custom_code()
        )

    @staticmethod
    def _contains_known_attack_signatures(path):
        unquoted = unquote(path)
        unescaped = unquoted.encode("utf-8").decode("unicode-escape")

        try:
            realpath = os.path.realpath(unescaped).lower().rstrip("/")
        except ValueError as e:
            return str(e) == "embedded null byte"
        except TypeError as e:
            return (
                "NUL" in str(e)
                or "null byte" in str(e)
                or str(e) == "embedded NUL character"
            )
        except Exception as e:
            return "null byte" in str(e).lower()

        return any(
            [
                bypass_markers.lower().rstrip("/") in realpath
                for bypass_markers in PathTraversalSinkFeatures.KNOWN_SECURITY_BYPASS_MARKERS
            ]
        )

    @staticmethod
    def _is_system_file(path):
        if not path:
            return False
        unquoted = unquote(path)
        realpath = os.path.realpath(unquoted).lower().rstrip("/")

        return any(
            [
                sys_file.lower().rstrip("/") in realpath
                for sys_file in PathTraversalSinkFeatures.SYSTEM_FILES
            ]
        )



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
from contrast.extern.six import PY3


class HardcodedKeyHolder(object):
    # Should be found
    STR_KEY = "foo"
    BLAH_KEY = b"foo" if PY3 else bytearray("foo")
    BLAH_AES = b"foo" if PY3 else bytearray("foo")
    DES_BLAH = b"foo" if PY3 else bytearray("foo")
    IVASDF = b"foo" if PY3 else bytearray("foo")
    BLAH_SECRET = b"foo" if PY3 else bytearray("foo")
    KEY_STRING = b"foo" if PY3 else "foo"
    VALUE_PREFIX = "__foo"

    # Should not be found - values
    KEY_INT = 12345
    KEY_EMPTY = []
    KEY_NON_BYTES = ["asdf"]
    CONTENT_CODES = [200, 300, 400, 5000]
    __KEY_PREFIX = "foo"
    KEY_REASSIGN = BLAH_KEY

    os.environ.setdefault("KEY_FUNCTION", "foo")
    KEY_FUNCTION = os.environ.get("KEY_FUNCTION")



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

import contrast
from contrast.extern.wrapt import register_post_import_hook
from contrast.agent.assess.policy.analysis import analyze
from contrast.agent.policy import patch_manager
from contrast.agent.policy.applicator import apply_assess_patch
from contrast.agent.policy.loader import Policy
from contrast.utils.decorators import fail_safely
from contrast.utils.patch_utils import patch_cls_or_instance


@fail_safely("Failed to apply policy to new session class")
def _apply_policy(session_cls):
    policy = Policy()

    for patch_policy in policy.policy_by_module["pyramid.session"]:
        if patch_policy.class_name == "CookieSession":
            apply_assess_patch(session_cls, patch_policy)


@fail_safely("Failed to apply assess policy for BaseCookieSessionFactory")
def _apply_assess(result, args, kwargs):
    context = contrast.CS__CONTEXT_TRACKER.current()

    policy = Policy()
    patch_policy = policy.policy_by_name["pyramid.session.BaseCookieSessionFactory"]

    analyze(context, patch_policy, result, args, kwargs)


def BaseCookieSessionFactory(orig_func, patch_policy, *args, **kwargs):
    """
    BaseCookieSessionFactory is a function that returns a new CookieSession class

    Since we can't instrument the new class directly using normal policy machinery, we
    instead apply our policy on-demand to the newly created class.
    """
    session_cls = None

    try:
        session_cls = orig_func(*args, **kwargs)
        _apply_policy(session_cls)
    finally:
        _apply_assess(session_cls, args, kwargs)

    return session_cls


def patch_pyramid(pyramid_session):
    patch_cls_or_instance(
        pyramid_session, "BaseCookieSessionFactory", BaseCookieSessionFactory
    )


def register_patches():
    register_post_import_hook(patch_pyramid, "pyramid.session")


def reverse_patches():
    pyramid_session = sys.modules.get("pyramid.session")
    if not pyramid_session:
        return

    patch_manager.reverse_patches_by_owner(pyramid_session)



from __future__ import print_function
from pip._vendor.pkg_resources import get_distribution

from contrast.agent.assess.string_tracker import StringTracker
from contrast.utils.context_tracker import ContextTracker

from contrast.extern.six import PY2


__version__ = get_distribution("contrast-agent").version

CS__CONTEXT_TRACKER = ContextTracker()

STRING_TRACKER = StringTracker()


# --- import aliases ---

from contrast.agent.assess.utils import get_properties  # noqa

if PY2:
    from contrast.extern import pathlib2 as pathlib
else:
    import pathlib  # noqa



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from os import path
import pytest
import six  # pylint: disable=did-not-import-extern
import sys

from contrast.agent.policy import constants
from contrast.agent.assess.utils import get_properties
from contrast.api.dtm_pb2 import TraceEvent
from contrast.test import mocks
from contrast.test.contract.events import (
    assert_propagation_event,
    assert_source_event,
    assert_trigger_event,
    check_event,
)
from contrast.test.contract.findings import (
    validate_finding_in_context,
    validate_source_finding_in_context,
    validate_header_sources,
    validate_cookies_sources,
)

from contrast.test.library_analysis import (
    assert_files_loaded,
    import_sample_package_onefile,
    TEST_MODULES_RELATIVE_IMPORT_DIST,
    TEST_MODULES_NAMESPACE_DIST,
    TEST_MODULE_ONEFILE,
    TEST_MODULES_MULTIPLE_TLDS_DIST,
)

from contrast.test.helper import python3_only


SOURCE_OPTIONS_MAP = {
    "args": "QUERYSTRING",
    "base_url": "URI",
    "full_path": "URI",
    "referer_header": "HEADER",
    "host": "URI",
    "host_url": "URI",
    "path": "URI",
    "query_string": "QUERYSTRING",
    "scheme": "OTHER",
    "url": "URI",
    "url_root": "URI",
    "values": "PARAMETER",
    "values_get_item": "PARAMETER",
}
# REMOTE_ADDR does not appear to be present in environ in Py27
SOURCE_OPTIONS_MAP.update({"remote_addr": "URI"} if six.PY3 else {})


POST_OPTIONS_MAP = {
    "files": "MULTIPART_CONTENT_DATA",
    "form": "MULTIPART_FORM_DATA",
    "wsgi.input": "BODY",
}

MULTIDICT_GET_OPTIONS_MAP = {
    "items": "QUERYSTRING",  # args.items()
    "lists": "QUERYSTRING",  # args.lists()
    "listvalues": "QUERYSTRING",  # args.listvalues()
    "values": "QUERYSTRING",  # args.values()
}

MULTIDICT_POST_OPTIONS_MAP = {
    "items": "MULTIPART_FORM_DATA",  # form.items()
    "lists": "MULTIPART_FORM_DATA",  # form.lists()
    "listvalues": "MULTIPART_FORM_DATA",  # form.listvalues()
    "values": "MULTIPART_FORM_DATA",  # form.values()
}


ALL_OPTIONS_MAP = {}
ALL_OPTIONS_MAP.update(SOURCE_OPTIONS_MAP)
ALL_OPTIONS_MAP.update(POST_OPTIONS_MAP)


SOURCE_OPTIONS = tuple(SOURCE_OPTIONS_MAP.keys())
POST_OPTIONS = tuple(POST_OPTIONS_MAP.keys())


def assert_flask_sqli_finding_events(finding, source_class_name):
    assert len(finding.events) == 16 if sys.version_info[:2] == (3, 5) else 17

    event_idx = 0
    check_event(
        finding.events[event_idx],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(constants.CREATION_TYPE),
        class_name=source_class_name,
        method_name="QUERY_STRING",
        source_types=["QUERYSTRING"],
        first_parent=None,
    )
    event_idx += 1
    check_event(
        finding.events[event_idx],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.OBJECT_KEY + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="str",
        method_name="encode",
        source_types=[],
        first_parent=finding.events[event_idx - 1],
    )
    event_idx += 1
    check_event(
        finding.events[event_idx],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.OBJECT_KEY + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="str",
        method_name="split",
        source_types=[],
        first_parent=finding.events[event_idx - 1],
    )
    event_idx += 2 if sys.version_info[:2] == (3, 5) else 1
    check_event(
        finding.events[event_idx],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.ALL_TYPE + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="str",
        method_name="replace",
        source_types=[],
        first_parent=finding.events[event_idx - 1],
    )
    event_idx += 2 if sys.version_info[:2] == (3, 5) else 1
    check_event(
        finding.events[event_idx],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.ALL_TYPE + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="str",
        method_name="CAST",
        source_types=[],
        first_parent=finding.events[event_idx - 1],
    )
    event_idx += 1
    check_event(
        finding.events[event_idx],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.ALL_TYPE + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="str",
        method_name="CAST",
        source_types=[],
        first_parent=finding.events[event_idx - 1],
    )
    event_idx += 1
    check_event(
        finding.events[event_idx],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.OBJECT_KEY + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="str",
        method_name="decode",
        source_types=[],
        first_parent=finding.events[event_idx - 1],
    )
    event_idx += 4
    check_event(
        finding.events[event_idx],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.ALL_TYPE + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="str",
        method_name="concat",
        source_types=[],
        first_parent=finding.events[event_idx - 1],
    )
    event_idx += 1
    check_event(
        finding.events[event_idx],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.ALL_TYPE + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="str",
        method_name="concat",
        source_types=[],
        first_parent=finding.events[event_idx - 1],
    )

    # There are a few f-string propagation events here that previously were
    # omitted because they have no effect (i.e. the resulting string is
    # identical to the input). They are now showing up because we had to make a
    # fix to join propagation, which affects f-strings as well.

    event_idx += 1 if sys.version_info[:2] == (3, 5) else 4
    check_event(
        finding.events[event_idx],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.ALL_TYPE + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="str",
        method_name="CAST",
        source_types=[],
        # TODO: PYT-1448 missing parent ids with CAST propagation
        first_parent=None,
    )
    event_idx += 1
    check_event(
        finding.events[event_idx],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(constants.TRIGGER_TYPE),
        class_name="sqlite3.Cursor",
        method_name="execute",
        source_types=[],
        first_parent=finding.events[event_idx - 1],
    )


class FlaskTestLibraryAnalysisBuiltinImport(object):
    def test_relative_import_hook(self):
        response = self.client.get(
            "/import_package_with_relative_imports/", {"rm_sys_mod_entries": 1}
        )

        assert response.status_code == 200

        assert_files_loaded(
            self.request_context.activity, TEST_MODULES_RELATIVE_IMPORT_DIST
        )

    def test_namespace_sample_package_import_hook(self):
        response = self.client.get(
            "/import_namespace_package/", {"rm_sys_mod_entries": 1}
        )

        assert response.status_code == 200

        assert_files_loaded(self.request_context.activity, TEST_MODULES_NAMESPACE_DIST)

    def test_module_import_onefile(self):
        response = self.client.get(
            "/import_package_onefile/", {"rm_sys_mod_entries": 1}
        )

        assert response.status_code == 200

        assert_files_loaded(self.request_context.activity, TEST_MODULE_ONEFILE)

    def test_sample_module_multiple_tlds(self):
        response = self.client.get(
            "/import_sample_dist_multiple_tlds/", {"rm_sys_mod_entries": 1}
        )

        assert response.status_code == 200

        assert_files_loaded(
            self.request_context.activity, TEST_MODULES_MULTIPLE_TLDS_DIST
        )

    def test_module_already_imported(self):
        import_sample_package_onefile()

        response = self.client.get("/import_package_onefile/")

        assert response.status_code == 200

        assert len(self.request_context.activity.library_usages) == 0


class FlaskAssessTestMixin(object):

    ATTACK_VALUE = "im an attack!"

    def assert_propagation_happened(self, apply_trigger):
        """
        This tests that apply_trigger was called with the "ret" argument
        that has an HTML_ENCODED tag, indicating that propagation did occur.
        """
        assert apply_trigger.called
        call_args = apply_trigger.call_args
        ret_arg = call_args[0][2]
        assert "HTML_ENCODED" in get_properties(ret_arg).tags

    @python3_only
    def test_non_utf8_decoding_cmdi(self):
        """
        param_val a non utf8 encoded string. This is the result of shift_jis encoding of
        '手袋'. This encoding is automatically done by chrome, not in our test client. So
        we manually provide the encoded bytes here.
        """
        trigger = "os-system"
        rule_id = "cmd-injection"
        # Couldn't rely on passing 手袋 in the request because the test client doesn't
        # use shift_jis to encode.
        param_val = b"\x8e\xe8\x91\xdc"
        # Since we fail to decode param_value using utf-8 we rely on replace error
        # handling to get this new string
        expected_value = "���"
        path = "/vulnpy/cmdi/{}".format(trigger)

        self.client.get(path, {"user_input": param_val})

        validate_finding_in_context(self.request_context, rule_id, expected_value)

    @pytest.mark.parametrize("method_name", ["get", "post"])
    def test_xss(self, method_name):
        user_input = "whatever"
        method = getattr(self.client, method_name)

        response = method("/vulnpy/xss/raw", {"user_input": user_input})

        assert response.status_code == 200
        validate_finding_in_context(self.request_context, "reflected-xss", user_input)

        # Post creates a redos finding in werkzeug parse_options_header
        num_findings = 2 if method_name == "post" else 1
        finding = self.request_context.activity.findings[num_findings - 1]

        trigger_event = finding.events[-1]
        assert trigger_event.signature.class_name == "flask.app.Flask"
        assert trigger_event.signature.method_name == "wsgi_app"

    @pytest.mark.parametrize("user_input", ["something <> dangerous", "something safe"])
    @pytest.mark.parametrize("sanitizer", ["markupsafe", "html"])
    @mocks.apply_trigger
    def test_sanitized_xss(self, apply_trigger, user_input, sanitizer):
        response = self.client.post(
            "/{}-sanitized-xss?user_input={}".format(sanitizer, user_input)
        )

        assert response.status_code == 200

        self.assert_propagation_happened(apply_trigger)
        assert len(self.request_context.activity.findings) == 0

    @pytest.mark.parametrize("source", SOURCE_OPTIONS)
    def test_all_get_sources(self, source):
        self.client.set_cookie("user_input", "attack_cookie")
        self.client.get(
            "/dynamic-sources/",
            {"user_input": self.ATTACK_VALUE, "source": source},
            headers={"Referer": "www.python.org"},
            extra_environ={"REMOTE_ADDR": "localhost"},
        )

        validate_source_finding_in_context(
            self.request_context, "reflected-xss", source, ALL_OPTIONS_MAP
        )

    @pytest.mark.parametrize("source", SOURCE_OPTIONS + POST_OPTIONS)
    def test_all_post_sources(self, source):
        self.client.set_cookie("user_input", "attack_cookie")
        file_path = path.join(self.app.root_path, "..", "..", "data", "testfile.txt")
        self.client.post(
            "/dynamic-sources/?user_input={}&source={}".format(
                self.ATTACK_VALUE, source
            ),
            {"user_input": self.ATTACK_VALUE},
            headers={"Referer": "www.python.org"},
            upload_files=[("file_upload", file_path)],
            extra_environ={"REMOTE_ADDR": "localhost"},
        )

        validate_source_finding_in_context(
            self.request_context, "reflected-xss", source, ALL_OPTIONS_MAP
        )

    @pytest.mark.parametrize("source", MULTIDICT_GET_OPTIONS_MAP.keys())
    def test_all_get_multidict(self, source):
        self.client.get(
            "/multidict-sources",
            {"user_input": self.ATTACK_VALUE, "source": source},
            headers={"user_input": self.ATTACK_VALUE},
        )

        validate_source_finding_in_context(
            self.request_context, "reflected-xss", source, MULTIDICT_GET_OPTIONS_MAP
        )

    @pytest.mark.parametrize("source", MULTIDICT_POST_OPTIONS_MAP.keys())
    def test_all_post_multidict(self, source):
        self.client.post(
            "/multidict-sources?source={}".format(source),
            {"user_input": self.ATTACK_VALUE},
            headers={"user_input": self.ATTACK_VALUE},
        )

        validate_source_finding_in_context(
            self.request_context, "reflected-xss", source, MULTIDICT_POST_OPTIONS_MAP
        )

    def test_cookie_get(self):
        """Cookies shouldn't trigger xss, so we need a separate cookie source test"""
        self.client.set_cookie("user_input", "attack_cookie")
        self.client.get("/cookie-source")
        validate_cookies_sources(self.request_context.activity.findings)

    def test_cookie_post(self):
        """Cookies shouldn't trigger xss, so we need a separate cookie source test"""
        self.client.set_cookie("user_input", "attack_cookie")
        self.client.post("/cookie-source")
        validate_cookies_sources(self.request_context.activity.findings)

    @pytest.mark.parametrize(
        "route,source_name",
        [("/header-source/", "header"), ("/header-key-source/", "header_key")],
    )
    @pytest.mark.parametrize("method_name", ["get", "post"])
    def test_non_xss_sources(self, method_name, route, source_name):
        method = getattr(self.client, method_name)
        method(route, headers={"Test-Header": "whatever"})

        validate_header_sources(self.request_context.activity.findings, source_name)

    @pytest.mark.parametrize("method_name", ("get", "post"))
    def test_http_method_not_source(self, method_name):
        getattr(self.client, method_name)("/method-source/")
        assert len(self.request_context.activity.findings) == 0

    def test_sqli_sqlalchemy(self):
        param_val = "doesnt matter"
        self.client.get("/sqli/", {"user_input": param_val})

        validate_finding_in_context(self.request_context, "sql-injection", param_val)
        # the exact event sequence is different for PY2/PY3
        if six.PY3:
            assert_flask_sqli_finding_events(
                self.request_context.activity.findings[0], "wsgi.environ"
            )

    @pytest.mark.parametrize("with_kwarg", [False, True])
    def test_unvalidated_redirect(self, with_kwarg):
        redirect_route = "/cmdi"
        self.client.get(
            "/unvalidated-redirect",
            {"user_input": redirect_route, "with_kwarg": with_kwarg},
        )

        validate_finding_in_context(
            self.request_context, "unvalidated-redirect", redirect_route,
        )

    @pytest.mark.parametrize("setdefault", [False, True])
    def test_trust_boundary_violation(self, setdefault):
        user_input = "hello"
        self.client.get(
            "/trust-boundary-violation",
            {"user_input": user_input, "setdefault": setdefault},
        )

        validate_finding_in_context(
            self.request_context, "trust-boundary-violation", user_input, trigger_arg=1
        )

    @python3_only
    def test_pathlib_trigger(self):
        """
        Sanity check that ensures pathlib and os patches are applied in the right order
        """
        user_input = "/foo/bar.txt"
        self.client.get("/pathlib-path-traversal", {"user_input": user_input})

        validate_finding_in_context(
            self.request_context, "path-traversal", user_input, trigger_arg=0
        )

    @pytest.mark.xfail(reason="TODO: PYT-1464 clean up source events")
    def test_simple_events_num_events(self):
        param_val = "doesnt matter"
        self.client.get("/simple-events-single-source/", {"user_input": param_val})

        assert len(self.request_context.activity.findings) == 1
        assert len(self.request_context.activity.findings[0].events) == 2

    def test_simple_events(self):
        param_val = "doesnt matter"
        self.client.get("/simple-events-single-source/", {"user_input": param_val})

        assert len(self.request_context.activity.findings) == 1

        source_event = self.request_context.activity.findings[0].events[-2]
        trigger_event = self.request_context.activity.findings[0].events[-1]

        assert_source_event(
            source_event, "werkzeug.datastructures.TypeConversionDict", "get"
        )
        assert_trigger_event(trigger_event, "os", "utime")

    @pytest.mark.xfail(reason="TODO: PYT-1464 clean up source events")
    def test_simple_events_two_sources_num_events(self):
        self.client.get("/simple-events-two-sources/", dict(foo="foo", bar="bar"))

        assert len(self.request_context.activity.findings) == 1
        assert len(self.request_context.activity.findings[0].events) == 4

        # This should be the first event
        foo_source = self.request_context.activity.findings[0].events[-4]
        assert_source_event(
            foo_source, "werkzeug.datastructures.TypeConversionDict", "get"
        )

    def test_simple_events_two_sources(self):
        self.client.get("/simple-events-two-sources/", dict(foo="foo", bar="bar"))

        assert len(self.request_context.activity.findings) == 1

        # First event assertion is in previous test
        bar_source = self.request_context.activity.findings[0].events[-3]
        prop_event = self.request_context.activity.findings[0].events[-2]
        trigger_event = self.request_context.activity.findings[0].events[-1]

        assert_source_event(
            bar_source, "werkzeug.datastructures.TypeConversionDict", "get"
        )
        assert_propagation_event(prop_event, "str", "concat")
        assert_trigger_event(trigger_event, "os", "utime")

    @pytest.mark.xfail(reason="TODO: PYT-1464 clean up source events")
    def test_simple_events_same_source_unused_num_events(self):
        user_input = "doesnt matter"
        self.client.get(
            "/simple-events-same-source-unused/", dict(user_input=user_input)
        )

        assert len(self.request_context.activity.findings) == 1
        assert len(self.request_context.activity.findings[0].events) == 2

    def test_simple_events_same_source_unused(self):
        param_val = "doesnt matter"
        self.client.get("/simple-events-same-source-unused/", {"user_input": param_val})

        assert len(self.request_context.activity.findings) == 1

        source_event = self.request_context.activity.findings[0].events[-2]
        trigger_event = self.request_context.activity.findings[0].events[-1]

        assert_source_event(
            source_event, "werkzeug.datastructures.TypeConversionDict", "get"
        )
        assert_trigger_event(trigger_event, "os", "utime")

    @pytest.mark.xfail(reason="TODO: PYT-1464 clean up source events")
    def test_simple_events_same_source_used_twice_num_events(self):
        user_input = "doesnt matter"
        self.client.get(
            "/simple-events-same-source-used-twice/", dict(user_input=user_input)
        )

        assert len(self.request_context.activity.findings) == 1
        assert len(self.request_context.activity.findings[0].events) == 4

        # This should be the first event
        first_source = self.request_context.activity.findings[0].events[-4]
        assert_source_event(
            first_source, "werkzeug.datastructures.TypeConversionDict", "get"
        )

    def test_simple_events_same_source_used_twice(self):
        user_input = "doesnt matter"
        self.client.get(
            "/simple-events-same-source-used-twice/", dict(user_input=user_input)
        )

        assert len(self.request_context.activity.findings) == 1

        # First event assertion is in previous test
        # TODO: PYT-1464 If it's not possible to come up with a solution that *doesn't*
        # involve omitting this particular event, then delete this assertion. Since the
        # sources are actually identical, I think it would be reasonable if only one
        # source event shows up for this trace, although ideally we can keep both.
        second_source = self.request_context.activity.findings[0].events[-3]
        prop_event = self.request_context.activity.findings[0].events[-2]
        trigger_event = self.request_context.activity.findings[0].events[-1]

        assert_source_event(
            second_source, "werkzeug.datastructures.TypeConversionDict", "get"
        )
        assert_propagation_event(prop_event, "str", "concat")
        assert_trigger_event(trigger_event, "os", "utime")

    @pytest.mark.xfail(reason="TODO: PYT-1464 clean up source events")
    def test_simple_events_two_sources_multiple_propagation_num_events(self):
        self.client.get(
            "/simple-events-two-sources-multiple-propagators/",
            dict(foo="foo", bar="bar"),
        )

        assert len(self.request_context.activity.findings) == 1
        assert len(self.request_context.activity.findings[0].events) == 7

    def test_simple_events_two_sources_multiple_propagation(self):
        self.client.get(
            "/simple-events-two-sources-multiple-propagators/",
            dict(foo="foo", bar="bar"),
        )

        assert len(self.request_context.activity.findings) == 1

        events = self.request_context.activity.findings[0].events

        # TODO: PYT-1464 enable commented out event assertions
        # foo = request.args.get("foo")
        # assert_source_event(
        #     events[-7], "werkzeug.datastructures.TypeConversionDict", "get"
        # )
        # # foo = foo + "something"
        # assert_propagation_event(events[-6], "str", "concat")
        # # foo = foo.capitalize()
        # assert_propagation_event(events[-5], "str", "capitalize", a2r=False)

        # bar = request.args.get("bar")
        assert_source_event(
            events[-4], "werkzeug.datastructures.TypeConversionDict", "get"
        )
        # bar = "whatever" + bar
        assert_propagation_event(events[-3], "str", "concat")
        # result = foo + bar
        assert_propagation_event(events[-2], "str", "concat")
        # os.utime(result)
        assert_trigger_event(events[-1], "os", "utime")

    @pytest.mark.xfail(reason="TODO: PYT-1464 clean up source events")
    def test_simple_events_single_source_circular_usage_num_events(self):
        user_input = "doesnt matter"
        self.client.get(
            "/simple-events-single-source-circular-usage", dict(user_input=user_input)
        )

        assert len(self.request_context.activity.findings) == 1
        # TODO: PYT-1464 there's an argument for only 4 events here (i.e. the source
        # event only appears once, but I think that 5 is more accurate if possible
        assert len(self.request_context.activity.findings[0].events) == 5

        events = self.request_context.activity.findings[0].events

        assert_source_event(
            events[-5], "werkzeug.datastructures.TypeConversionDict", "get"
        )
        assert_propagation_event(events[-4], "str", "capitalize", a2r=False)

    def test_simple_events_single_source_circular_usage(self):
        user_input = "doesnt matter"
        self.client.get(
            "/simple-events-single-source-circular-usage", dict(user_input=user_input)
        )

        assert len(self.request_context.activity.findings) == 1

        events = self.request_context.activity.findings[0].events

        assert_source_event(
            events[-3], "werkzeug.datastructures.TypeConversionDict", "get"
        )
        assert_propagation_event(events[-2], "str", "concat")
        assert_trigger_event(events[-1], "os", "utime")



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import division
import time


class Timer(object):
    """
    Utility class to store timing by a key

    Example:
        t = Timer()

        set_start('bob')
        something()
        set_end('body')

    """

    def __init__(self, start_time=None):
        if start_time is None:
            start_time = time.time()

        self.start_at = start_time
        self.start_ms = int(self.start_at * 1000)
        self.events = dict()

        self.result = None

    def set_start(self, name):
        self.events[name] = [Timer.now_ms(), None]

    def set_end(self, name):
        self.events[name][1] = Timer.now_ms()

    @staticmethod
    def now_ms():
        return int(time.time() * 1000)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Definitions of blueprints that can be used in flask test apps
"""
import html
import io
import os
from os import path

from flask import Blueprint, render_template, request, redirect
from flask.sessions import SecureCookieSession
import contrast
from contrast.test.library_analysis import (
    remove_sys_module_entries,
    import_with_relative_imports,
    import_namespace_package,
    import_sample_package_onefile,
    import_multiple_tlds,
)

FRAMEWORK_TEST_DIR = path.join(path.dirname(contrast.__file__), "..")
TEMPLATE_DIR = path.join(FRAMEWORK_TEST_DIR, "templates", "flask")
router = Blueprint("contrast", __name__, template_folder=TEMPLATE_DIR)


@router.route(
    "/import_package_with_relative_imports", methods=["GET"], strict_slashes=False
)
def import_package_with_relative_imports():
    rm_modules = request.args.get("rm_sys_mod_entries", None)

    if rm_modules:
        remove_sys_module_entries()

    import_with_relative_imports()

    return ""


@router.route("/import_namespace_package", methods=["GET"], strict_slashes=False)
def import_sample_namespace_package():
    # Deleting top level namespace modules breaks them in py2.
    # Namespace name is cached in sys modules, but shows up
    # as a builtin module in py2.
    import_namespace_package()

    return ""


@router.route("/import_package_onefile", methods=["GET"], strict_slashes=False)
def import_package_onefile():
    rm_modules = request.args.get("rm_sys_mod_entries", None)

    if rm_modules:
        remove_sys_module_entries()

    import_sample_package_onefile()

    return ""


@router.route(
    "/import_sample_dist_multiple_tlds", methods=["GET"], strict_slashes=False
)
def import_sample_dist_multiple_tlds():
    rm_modules = request.args.get("rm_sys_mod_entries", None)

    if rm_modules:
        remove_sys_module_entries()

    import_multiple_tlds()

    return ""


@router.route(
    "/markupsafe-sanitized-xss", methods=["GET", "POST"], strict_slashes=False
)
def markupsafe_sanitized_xss():
    """
    This is a test for Markupsafe.escape as it gets called by the template engine.
    """
    user_input = request.args.get("user_input")
    ret = render_template("sanitized_xss.html", user_input=user_input)
    return ret


@router.route("/html-sanitized-xss", methods=["GET", "POST"], strict_slashes=False)
def html_sanitized_xss():
    """
    This is a test for Markupsafe.escape as it gets called by the template engine.
    """
    user_input = request.args.get("user_input")
    ret = html.escape(user_input)
    return ret


@router.route("/dynamic-sources", methods=["GET", "POST"], strict_slashes=False)
def dynamic_sources():
    source = request.args.get("source", "")
    user_input = ""

    if source == "args":
        user_input = request.args.get("user_input")
    elif source == "base_url":
        user_input = request.base_url
    elif source == "referer_header":
        user_input = request.headers.get("Referer")
    elif source == "host":
        user_input = request.host
    elif source == "host_url":
        user_input = request.host_url
    elif source == "files" and request.method == "POST":
        stream = request.files.get("file_upload")
        user_input = stream.read()
    elif source == "form" and request.method == "POST":
        user_input = request.form.get("user_input")
    elif source == "full_path":
        user_input = request.full_path
    elif source == "path":
        user_input = request.path
    elif source == "query_string":
        user_input = request.query_string
    elif source == "remote_addr":
        user_input = request.remote_addr
    elif source == "scheme":
        user_input = request.scheme
    elif source == "url":
        user_input = request.url
    elif source == "url_root":
        user_input = request.url_root
    elif source == "values":
        user_input = request.values.get("user_input")
    elif source == "values_get_item":
        user_input = request.values["user_input"]
    elif source == "wsgi.input":
        user_input = request.environ["wsgi.input"].read()
        # restore wsgi.input so that it can still be read later if needed
        request.environ["wsgi.input"] = io.BytesIO(user_input)

    return render_template("xss.html", user_input=user_input)


@router.route("/cookie-source", methods=["GET", "POST"], strict_slashes=False)
def cookie_source():
    user_input = request.cookies["user_input"]
    os.system("echo {}".format(user_input))
    return render_template("xss.html", user_input=user_input)


@router.route("/header-source", methods=["GET", "POST"], strict_slashes=False)
def header_source():
    user_input = request.headers.get("Test-Header")
    os.system("echo {}".format(user_input))
    return render_template("xss.html", user_input=user_input)


@router.route("/header-key-source", methods=["GET", "POST"], strict_slashes=False)
def header_key_source():
    header_keys = list(request.headers.keys())
    # Make sure we grab a header that we know will not come from another source, so it
    # will not have any other tags. For example, the 'Host' header has source type URI,
    # and so gets tagged with CROSS_SITE, so we don't want to test with that one.
    user_input = header_keys[header_keys.index("Test-Header")]
    os.system("echo {}".format(user_input))
    return render_template("xss.html", user_input=user_input)


@router.route("/method-source", methods=["GET", "POST"], strict_slashes=False)
def method_source():
    user_input = request.method
    os.system("echo {}".format(user_input))
    return render_template("xss.html", user_input=user_input)


@router.route("/multidict-sources", methods=["GET", "POST"], strict_slashes=False)
def multidict_sources():
    source = request.args.get("source", "")
    user_input = ""

    if request.method == "GET":
        if source == "items":
            user_input = [
                x for x in list(request.args.items()) if x[0] == "user_input"
            ][0][1]
        elif source == "lists":
            user_input = [
                x for x in list(request.args.lists()) if x[0] == "user_input"
            ][0][1][0]
        elif source == "listvalues":
            user_input = list(request.args.listvalues())[0][0]
        elif source == "values":
            user_input = list(request.args.values())[0]
    elif request.method == "POST":
        if source == "items":
            user_input = [
                x for x in list(request.form.items()) if x[0] == "user_input"
            ][0][1]
        elif source == "lists":
            user_input = [
                x for x in list(request.form.lists()) if x[0] == "user_input"
            ][0][1][0]
        elif source == "listvalues":
            user_input = list(request.form.listvalues())[0][0]
        elif source == "values":
            user_input = list(request.form.values())[0]

    return render_template("xss.html", user_input=user_input)


@router.route("/unvalidated-redirect", methods=["GET", "POST"], strict_slashes=False)
def unvalidated_redirect():
    user_input = request.args.get("user_input")
    if request.args.get("with_kwarg", str(False)) == str(True):
        return redirect(location=user_input)
    return redirect(user_input)


@router.route("/trust-boundary-violation", methods=["GET"], strict_slashes=False)
def trust_boundary_violation():
    user_input = request.args.get("user_input")
    session = SecureCookieSession()

    if request.args.get("setdefault") == "True":
        session.setdefault("user_input", user_input)
    else:
        session["user_input"] = user_input

    return "<p>Trust boundary violation</p>"


@router.route("/pathlib-path-traversal", methods=["GET"], strict_slashes=False)
def pathlib_path_traversal():
    import pathlib

    user_input = request.args.get("user_input")

    path = pathlib.Path(user_input)

    try:
        # This method is known to use the accessor and so it enables us to test whether
        # the patches are applied in the correct order (i.e. os before pathlib)
        path.rmdir()
    except Exception:
        pass

    return "<p>Pathlib path traversal<p>"


@router.route("/simple-events-single-source", methods=["GET"], strict_slashes=False)
def simple_events_single_source():
    """View with a single source event and single trigger event"""
    # This *should* be the first event
    user_input = request.args.get("user_input")

    try:
        # This *should* be the second event
        os.utime(user_input)
    except Exception:
        pass

    return "<p>Simple events single source</p>"


@router.route("/simple-events-two-sources", methods=["GET"], strict_slashes=False)
def simple_events_two_sources():
    """View with two source events, a propagation event, and single trigger event"""
    # This *should* be the first event
    foo = request.args.get("foo")
    # This *should* be the second event
    bar = request.args.get("bar")

    # This *should* be the third event
    result = foo + bar

    try:
        # This *should* be the fourth event
        os.utime(result)
    except Exception:
        pass

    return "<p>Simple events two sources</p>"


@router.route(
    "/simple-events-same-source-unused", methods=["GET"], strict_slashes=False
)
def simple_event_same_source_unused():
    """
    View with two source events using the same string

    One contributes to the finding and the other does not
    """
    # This *should* be the first event
    user_input = request.args.get("user_input")
    # This event should not appear in the trace at all even though the underlying
    # string is identical, which means we are actually appending another source event
    # here.
    unused = request.args.get("user_input")  # pylint: disable=unused-variable

    try:
        # This *should* be the second event
        os.utime(user_input)
    except Exception:
        pass

    return "<p>Simple events same source unused</p>"


@router.route(
    "/simple-events-same-source-used-twice", methods=["GET"], strict_slashes=False
)
def simple_event_same_source_used_twice():
    """
    View with two source events using the same string where both contribute to finding
    """
    # This *should* be the first event
    user_input = request.args.get("user_input")
    # This *should* be the second event, although I think it wouldn't be the worst
    # thing if this event had to be omitted...
    again = request.args.get("user_input")

    # This *should* be the third event
    result = user_input + again

    try:
        # This *should* be the fourth event
        os.utime(result)
    except Exception:
        pass

    return "<p>Simple events same source used twice</p>"


@router.route(
    "/simple-events-two-sources-multiple-propagators",
    methods=["GET"],
    strict_slashes=False,
)
def simple_events_two_sources_multiple_propagators():
    """View with two source events and several chains of propagation"""
    # This *should* be the first event
    foo = request.args.get("foo")
    # This *should* be the second event
    bar = request.args.get("bar")

    foo = foo + "something"
    foo = foo.capitalize()

    bar = "whatever" + bar

    result = foo + bar

    try:
        os.utime(result)
    except Exception:
        pass

    return "<p>Simple events two sources multiple propagators<p>"


@router.route(
    "/simple-events-single-source-circular-usage",
    methods=["GET"],
    strict_slashes=False,
)
def simple_events_single_source_circular_usage():
    """View with single source event where source string gets used circularly"""
    # This *should* be the first event
    user_input = request.args.get("user_input")

    result = user_input.capitalize()
    result = result + user_input

    try:
        os.utime(result)
    except Exception:
        pass

    return "<p>Simple events single source circular usage</p>"



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import io
import os
import sys
from collections import Counter

import mock
import pytest
from contrast.extern import six


python2_only = pytest.mark.skipif(sys.version_info[0] > 2, reason="Python 2 only")
python3_only = pytest.mark.skipif(sys.version_info[0] < 3, reason="Python 3 only")
skip_if_python39 = pytest.mark.skipif(
    sys.version_info[:2] == (3, 9), reason="Not supported on Python 3.9"
)


class SimpleFile(object):
    name = "my_file.txt"
    filename = "my_file.txt"
    file = io.BytesIO(b"hello world")


SIMPLE_REQUEST = {
    "HTTP_USER_AGENT": "",
    "wsgi.url_scheme": "http",
    "HTTP_HOST": "google.com",
    "PATH_INFO": "/bing",
    "HTTP_COOKIE": "",
    "QUERY_STRING": "",
    "REMOTE_ADDR": "127.0.0.1",
    "wsgi.input": io.BytesIO(b""),
    "wsgi.errors": io.BytesIO(b""),
    "REQUEST_METHOD": "GET",
    "SERVER_PORT": 8000,
    "SERVER_PROTOCOL": "HTTP/1.1",
}


def get_simple_request():
    return SIMPLE_REQUEST.copy()


def printable_vector(vector):
    if six.PY2:
        return vector.encode("utf-8")

    return vector


skipInCI = pytest.mark.skipif(
    os.environ.get("CI", "") == "true", reason="Bitbucket pipelines symlink"
)

mock_build_update_messages = mock.patch(
    "contrast.agent.service_client.build_update_message"
)

mock_send_messages = mock.patch("contrast.agent.service_client.send_messages")


def compare_two_lists(a, b):
    return Counter(a) == Counter(b)


SETTINGS_LOC = "contrast.agent.settings_state.SettingsState"
protect_enabled = mock.patch(SETTINGS_LOC + ".is_protect_enabled", return_value=True)
assess_enabled = mock.patch(SETTINGS_LOC + ".is_assess_enabled", return_value=True)
protect_disabled = mock.patch(SETTINGS_LOC + ".is_protect_enabled", return_value=False)
assess_disabled = mock.patch(SETTINGS_LOC + ".is_assess_enabled", return_value=False)


@pytest.fixture(autouse=True)
def disable_library_analysis(mocker):
    mocker.patch(
        "contrast.agent.middlewares.base_middleware.BaseMiddleware.initialize_libraries"
    )



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os

import django
import pytest

from contrast.test.framework_test.base_test import BaseContrastTest


@pytest.mark.django_db
class BaseDjangoTest(BaseContrastTest):
    @property
    def middleware_request_context(self):
        # TODO: PYT-1102
        if os.environ.get("USE_WSGI"):
            return "contrast.wsgi.middleware.RequestContext"
        return (
            "contrast.agent.middlewares.django_middleware.RequestContext"
            if django.VERSION >= (1, 10)
            else "contrast.agent.middlewares.legacy_django_middleware.RequestContext"
        )

    @property
    def application_module_name(self):
        return "app.wsgi"

    @property
    def application_attribute_name(self):
        return "application"



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern.wrapt import register_post_import_hook
import sys
import contrast
from contrast.agent.policy import patch_manager
from contrast.patches.databases import dbapi2
from contrast.utils import inventory_utils
from contrast.utils.patch_utils import patch_cls_or_instance

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


SQLITE3 = "sqlite3"
PYSQLITE2_DBAPI2 = "pysqlite2.dbapi2"
VENDOR = "SQLite3"


def patch_connect(module):
    def _connect(orig_func, patch_policy=None, *args, **kwargs):
        """Record DB inventory for SQLite3 connection"""
        try:
            context = contrast.CS__CONTEXT_TRACKER.current()
            if context is not None:
                # sqlite does not use a server, so no need for host/port
                database = kwargs.get("database") or (
                    args[0] if len(args) > 0 else "unknown"
                )
                db_inventory = dict(vendor=VENDOR, database=database)
                inventory_utils.append_db(context.activity, db_inventory)
        except Exception:
            logger.exception("Failed to add inventory for %s", VENDOR)

        return orig_func(*args, **kwargs)

    patch_cls_or_instance(module, "connect", _connect)


def instrument_sqlite_dbapi2(sqlite3):
    dbapi2.instrument_cursor(SQLITE3, sqlite3.Cursor)
    dbapi2.instrument_executescript(SQLITE3, sqlite3.Cursor)
    patch_connect(sqlite3)


def instrument_pysqlite2_dbapi2(pysqlite2):
    """Supports the older pysqlite module in Py2"""
    dbapi2.instrument_cursor(PYSQLITE2_DBAPI2, pysqlite2.Cursor)
    dbapi2.instrument_executescript(PYSQLITE2_DBAPI2, pysqlite2.Cursor)
    patch_connect(pysqlite2)


def register_patches():
    register_post_import_hook(instrument_sqlite_dbapi2, SQLITE3)
    register_post_import_hook(instrument_pysqlite2_dbapi2, PYSQLITE2_DBAPI2)


def reverse_patches():
    sqlite3 = sys.modules.get(SQLITE3)
    if sqlite3:
        patch_manager.reverse_patches_by_owner(sqlite3)
        patch_manager.reverse_patches_by_owner(sqlite3.Cursor)

    psqlite = sys.modules.get(PYSQLITE2_DBAPI2)
    if psqlite:
        patch_manager.reverse_patches_by_owner(psqlite)
        patch_manager.reverse_patches_by_owner(psqlite.Cursor)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern.six import string_types, binary_type

from contrast.agent import scope
from contrast.agent.policy.loader import Policy
from contrast.agent.assess.policy import trigger_policy
from contrast.utils.decorators import fail_safely


@fail_safely("Error running unsafe code execution assess rule")
def apply_rule(module_name, method_name, result, args, kwargs):
    if len(args) < 1:
        return

    if not isinstance(args[0], string_types + (binary_type, bytearray)):
        return

    policy = Policy()
    trigger_rule = policy.triggers["unsafe-code-execution"]

    trigger_nodes = trigger_rule.find_trigger_nodes(module_name, method_name)

    with scope.trigger_scope():
        trigger_policy.apply(trigger_rule, trigger_nodes, result, args, kwargs)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from pip._vendor import pkg_resources
from collections import namedtuple

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")

Version = namedtuple("Version", ["major", "minor", "patch"])
SUPPORTED_FRAMEWORKS = ["django", "falcon", "flask", "pylons", "pyramid"]
DEFAULT_FRAMEWORK = "WSGI"


class Framework(object):
    """
    A class to store information about the
    current web framework used for an application
    """

    def __init__(self):
        self._name = ""
        self.version = None
        self.set_info()

    @property
    def name(self):
        return self._name.capitalize()

    @property
    def name_lower(self):
        return self._name.lower()

    @property
    def full_version(self):
        return "{}.{}.{}".format(
            self.version.major, self.version.minor, self.version.patch
        )

    def set_info(self):
        framework = self.discover_framework()

        if framework:
            version = framework.version.split(".")
            patch = version[2] if len(version) > 2 else "0"
            self.version = Version(major=version[0], minor=version[1], patch=patch)
            self._name = framework.project_name
        else:
            logger.debug(
                "Did not find the current framework. Assuming it's %s.",
                DEFAULT_FRAMEWORK,
            )
            self._name = DEFAULT_FRAMEWORK
            self.version = Version(major="0", minor="0", patch="0")

    def discover_framework(self):
        """
        Except in the agent's own testing environment, the assumption here is that all environments
        using the agent will have only one supported framework.

        :return pkg_resources.DistInfoDistribution instance
        """
        for framework_name in SUPPORTED_FRAMEWORKS:
            try:
                return pkg_resources.get_distribution(framework_name)
            except Exception:
                pass

    def __repr__(self):
        return "{} {}".format(self.name, self.full_version)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import threading
import time
from contrast.agent import scope
from contrast.utils.decorators import fail_safely
from contrast.utils.timer import Timer

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class Heartbeat(threading.Thread):
    def __init__(self, settings):
        super(Heartbeat, self).__init__()
        self.daemon = True
        self.settings = settings
        self.stopped = False

        self.heartbeat_interval_ms = self.settings.config.get(
            "agent.polling.app_activity_ms", 30000
        )
        self.heartbeat_interval_sec = self.heartbeat_interval_ms / 1000

        # Agent should not ping too frequently
        if self.heartbeat_interval_ms < 10000:
            self.heartbeat_interval_ms = 10000
            self.heartbeat_interval_sec = 10

    def get_sleep_time(self):
        return self.heartbeat_interval_sec

    def run(self):
        # Ensure the heartbeat thread runs in scope because it is initialized
        # before our thread.start patch is applied.
        with scope.contrast_scope():
            logger.debug("Establishing heartbeat")

            while not self.stopped and self.settings.is_agent_config_enabled():
                self.send_heartbeat()
                time.sleep(self.get_sleep_time())

    @fail_safely("Error sending a heartbeat message")
    def send_heartbeat(self):
        from contrast.agent import service_client

        heartbeat_interval_ms = self.heartbeat_interval_ms

        # only send after we have not updated after the last interval time set in the config
        if (
            self.settings.last_update is not None
            and (Timer.now_ms() - self.settings.last_update) > heartbeat_interval_ms
        ):
            service_client.send_heartbeat_message()



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""Central location for commonly used mocks"""
import mock

from contrast.api import send_finding
from contrast.agent.assess.policy import trigger_policy
from contrast.api.dtm_pb2 import UserInput
from contrast.api.settings_pb2 import InputAnalysis, InputAnalysisResult

apply_trigger = mock.patch(
    "contrast.agent.assess.policy.trigger_policy.apply",
    side_effect=trigger_policy.apply,
)

send_finding = mock.patch(
    "contrast.agent.assess.rules.trigger_rule.send_finding", side_effect=send_finding
)

send_finding_static = mock.patch(
    "contrast.agent.assess.rules.static_rule.send_finding", side_effect=send_finding
)


def nosqli_input_analysis():
    """
    Return input analysis results as if Speedracer had processed the input.
    Note that the score_level may be different than what real Speedracer returns.
    """
    result = InputAnalysisResult(
        rule_id="nosql-injection",
        value='{"title": "Record One"}',
        path="user_input",
        key="user_input",
        input_type=UserInput.PARAMETER_VALUE,
        score_level=InputAnalysisResult.DEFINITEATTACK,
        ids=[],
    )

    return InputAnalysis(results=[result])


nosqli_input_analysis_mock = mock.patch(
    "contrast.agent.middlewares.base_middleware.get_input_analysis",
    return_value=nosqli_input_analysis(),
)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os

from contrast.extern.six import string_types
from contrast.agent.protect.mixins.path_traversal_sink_features import (
    PathTraversalSinkFeatures,
)
from contrast.agent.protect.rule.base_rule import BaseRule

from contrast.utils.decorators import fail_quietly

PARENT_CHECK = ".."
SLASH = "/"
SAFE_PATHS = ["tmp", "public", "docs", "static", "template", "templates"]


class PathTraversal(BaseRule, PathTraversalSinkFeatures):
    NAME = "path-traversal"

    def __init__(self, settings):
        BaseRule.__init__(self, settings)

    def find_attack(self, candidate_string=None, **kwargs):
        """
        Finds the attacker in the original string if present
        """
        attack = super(PathTraversal, self).find_attack(candidate_string, **kwargs)
        if self.in_infilter:
            attack = self.check_sink_features(candidate_string, attack)

        return attack

    def build_sample(self, evaluation, path, **kwargs):
        sample = self.build_base_sample(evaluation)
        if path is not None:
            sample.path_traversal.path = path
        return sample

    @fail_quietly(
        "Failed to run path traversal skip_protect_analysis", return_value=False
    )
    def skip_protect_analysis(self, user_input, args, kwargs):
        write = possible_write(args, kwargs)
        if write:
            # any write is a risk so we should not skip analysis
            return False

        return not actionable_path(user_input)

    def infilter_kwargs(self, user_input, patch_policy):
        return dict(method=patch_policy.method_name)


def possible_write(args, kwargs):
    if "w" in kwargs.get("mode", ""):
        return True

    return len(args) > 1 and args[1] is not None and "w" in args[1]


def actionable_path(path):
    if not path or not isinstance(path, string_types):
        return False

    # moving up directory structure is a risk and hence actionable
    if path.find(PARENT_CHECK) > 1:
        return True

    if "/contrast/" in path or "/site-packages/" in path:
        return False

    if path.startswith(SLASH):
        for prefix in _safer_abs_paths():
            if path.startswith(prefix):
                return False
    else:
        for prefix in SAFE_PATHS:
            if path.startswith(prefix):
                return False

    return True


def _safer_abs_paths():
    pwd = os.getcwd()

    return ["{}/{}".format(pwd, item) for item in SAFE_PATHS] if pwd else []



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern.six import ensure_binary, iteritems
from contrast.agent import scope

from contrast.agent.request import Request
from contrast.agent.settings_state import SettingsState
from contrast.api.dtm_pb2 import Activity, ServerActivity, ObservedRoute
from contrast.utils.timer import Timer

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class RequestContext(object):
    def __init__(self, environ, body=None):

        logger.debug("Initializing Request Context")

        scope.enter_contrast_scope()

        self.timer = Timer()

        self.request = Request(environ, body)

        dtm = self.request.get_dtm()
        self.activity = Activity()
        self.activity.http_request.CopyFrom(dtm)

        self.server_activity = ServerActivity()

        self.speedracer_input_analysis = None
        self.do_not_track = False

        # to be populated with a RouteCoverage instance
        self.current_route = None

        self.observed_route = ObservedRoute()

        scope.exit_contrast_scope()

    @property
    def propagate_assess(self):
        # TODO: PYT-644 move this property of out this class?
        return SettingsState().is_assess_enabled() and not scope.in_scope()

    def extract_response_to_context(self, response):
        """
        Append response to request context and populate the HttpResponse DTM

        :response: Subclass of BaseResponseWrapper
        """
        self.response = response

        if not SettingsState().response_scanning_enabled:
            return

        self.activity.http_response.response_code = response.status_code

        # From the dtm for normalized_response_headers:
        #   Key is UPPERCASE_UNDERSCORE
        #
        #   Example: Content-Type: text/html; charset=utf-8
        #   "CONTENT_TYPE" => Content-Type,["text/html; charset=utf8"]
        for key, values in iteritems(response.headers.dict_of_lists()):
            normalized_key = key.upper().replace("-", "_")
            response_headers = self.activity.http_response.normalized_response_headers
            response_headers[normalized_key].key = key
            response_headers[normalized_key].values.extend(values)

        self.activity.http_response.response_body_binary = ensure_binary(
            response.body or ""
        )

    def get_xss_findings(self):
        """
        Return a list of Finding obj of rule_id reflected-xss, if any exist in Activity
        """
        return [
            finding
            for finding in self.activity.findings
            if finding.rule_id == "reflected-xss"
        ]



from collections import namedtuple
import os

import contrast.utils.configuration_utils as utils

from contrast.utils.loggers import DEFAULT_PROGNAME
from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


SPLIT_CWD = os.getcwd().split(os.sep)
APP_NAME = SPLIT_CWD[len(SPLIT_CWD) - 1]
ENV_PREFIX = "CONTRAST"

ConfigOption = namedtuple("ConfigOption", ["name", "default_value", "type_cast"])


DEFAULT_SERVICE_HOST = "127.0.0.1"
DEFAULT_SERVICE_PORT = 30555


class AgentConfig(object):
    def __init__(self):
        self._config = utils.load_yaml_config()

        if not self._config:
            logger.info("No YAML config found; using default settings")
            self._config = {}

        self._config = utils.flatten_config(self._config)

        self.override_configs()

        # fallback for no host or no port specified
        # socket will still take precedence if it exists
        self.check_for_service_config()

    def get(self, key, default_value=None):
        return self._config.get(key, default_value)

    def put(self, key, value):
        if not key or key not in self._config:
            logger.error("Invalid contrast config option: %s", key)
            return

        self._config[key] = value

    def log_config(self):
        """Log configuration values EXCEPT API information containing sensitive data"""
        config_no_api = {
            key: self._config[key] for key in self._config if "api" not in key
        }

        logger.info("Current Configuration", **config_no_api)

    def override_configs(self):
        """
        For each class representing different parts of the config (Agent, etc),
        take the current config and apply any overriding logic such as overriding with
        environment keys/values.

        At this time the order of precedence is:
            os.env > Config yaml

        Meaning that a config value defined in os.environ must be used instead of
        the same config defined in contrast_security.yaml

        Note that CLI args (sys.argv) are not supported. This may change if the
        agent becomes a runner.
        """
        builders = [Agent, Application, Assess, Inventory, Protect, Root, Server]

        for builder in builders:
            builder().build(self._config)

    def check_for_service_config(self):
        """
        Check for and log user-provided config values for connecting to SR.
        Socket config takes precedence if it is user-provided. Otherwise assign
        host/port values if either is not provided.
        """
        socket_key = "agent.service.socket"
        host_key = "agent.service.host"
        port_key = "agent.service.port"

        # Socket is set
        if self.get(socket_key):
            logger.info(
                "Connecting to the Contrast Service using a UnixSocket socket: %s",
                self.get(socket_key),
            )
            return

        # Host and Port are set
        if self.get(host_key) and self.get(port_key):
            logger.info(
                "Connecting to the Contrast Service using a TCP socket: host: %s, port: %s",
                self.get(host_key),
                self.get(port_key),
            )
            return

        # Something is not set

        if self.get(host_key):
            self.put(port_key, DEFAULT_SERVICE_PORT)
            msg = "{} is not set. Falling back to default TCP socket host: {}, port: {}".format(
                port_key, self.get(host_key), self.get(port_key)
            )
        elif self.get(port_key):
            self.put(host_key, DEFAULT_SERVICE_HOST)
            msg = "{} is not set. Falling back to default TCP socket host: {}, port: {}".format(
                host_key, self.get(host_key), self.get(port_key)
            )
        else:
            self.put(host_key, DEFAULT_SERVICE_HOST)
            self.put(port_key, DEFAULT_SERVICE_PORT)
            msg = "Neither {} nor the pair of {} and {} are set. Falling back to default TCP socket host: {}, port: {}.".format(
                socket_key, host_key, port_key, self.get(host_key), self.get(port_key),
            )

        msg_prefix = "Missing a required connection value to the Contrast Service."
        warn_msg = "{} {}".format(msg_prefix, msg)

        logger.warning(warn_msg)

    def get_session_id(self):
        return self.get("application.session_id", "")

    def get_session_metadata(self):
        return self.get("application.session_metadata", "")


class ConfigBuilder(object):
    TOP_LEVEL = ""

    def __init__(self):
        self.default_options = []

    def build(self, config):
        """
        Given a dict config, iterate over the default_options and check if
        the corresponding config key/value should be either :
        1. replaced by an existing env var
        2. keep existing config key/val but type-cast the value
        3. add a new key/default_value to the config

        :param config: dict config
        :return: None, config dict is updated pass by reference
        """
        for option_name, default_val, type_cast in self.default_options:

            dot_alt = self._dot_alternative(option_name)
            underscore_alt = self._underscore_alternative(option_name)

            env_override = utils.get_env_key(underscore_alt)
            if env_override:
                # replace the config value with the os.env value and apply type-cast
                env_override = type_cast(env_override)
                config[dot_alt] = env_override
                continue

            if dot_alt in config:
                # update the config value with a type-cast value
                val = config[dot_alt]
                val = type_cast(val)
                config[dot_alt] = val
                continue

            # add a new key/default_value to config
            config[dot_alt] = default_val

    def _underscore_alternative(self, key):
        return "__".join(
            [x for x in [ENV_PREFIX, self.TOP_LEVEL, key.replace(".", "__")] if x]
        ).upper()

    def _dot_alternative(self, key):
        return ".".join([self.TOP_LEVEL, key]) if self.TOP_LEVEL else key


class Agent(ConfigBuilder):
    TOP_LEVEL = "agent"

    def __init__(self):
        super(Agent, self).__init__()

        self.default_options = [
            # Some logger default values are handled by the logger
            ConfigOption(name="logger.level", default_value="", type_cast=str,),
            ConfigOption(name="logger.path", default_value="", type_cast=str),
            ConfigOption(
                name="logger.progname", default_value=DEFAULT_PROGNAME, type_cast=str,
            ),
            ConfigOption(
                name="python.enable_drf_response_analysis",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                name="python.enable_rep",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                name="polling.app_activity_ms", default_value=30000, type_cast=int
            ),
            ConfigOption(
                name="polling.server_activity_ms", default_value=30000, type_cast=int
            ),
            ConfigOption(
                name="polling.accumulated_rules_ms", default_value=30000, type_cast=int
            ),
            ConfigOption(
                name="service.enable", default_value=True, type_cast=utils.str_to_bool
            ),
            ConfigOption(name="service.host", default_value="", type_cast=str),
            ConfigOption(name="service.port", default_value="", type_cast=int),
            ConfigOption(name="service.socket", default_value="", type_cast=str),
        ]


class Application(ConfigBuilder):
    TOP_LEVEL = "application"

    def __init__(self):
        super(Application, self).__init__()

        self.default_options = [
            ConfigOption(name="code", default_value="", type_cast=str),
            ConfigOption(name="group", default_value="", type_cast=str),
            ConfigOption(name="metadata", default_value="", type_cast=str),
            ConfigOption(
                name="name", default_value=os.path.basename(os.getcwd()), type_cast=str
            ),
            ConfigOption(name="path", default_value="", type_cast=str),
            ConfigOption(name="tags", default_value="", type_cast=str),
            ConfigOption(name="version", default_value="", type_cast=str),
            ConfigOption(name="session_id", default_value="", type_cast=str),
            ConfigOption(name="session_metadata", default_value="", type_cast=str),
        ]


class Assess(ConfigBuilder):
    TOP_LEVEL = "assess"

    def __init__(self):
        super(Assess, self).__init__()

        self.default_options = [
            ConfigOption(
                name="enable", default_value=None, type_cast=utils.str_to_bool
            ),
            ConfigOption(
                name="enable_scan_response",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(name="sampling.baseline", default_value=5, type_cast=int),
            ConfigOption(
                name="sampling.enable", default_value=True, type_cast=utils.str_to_bool
            ),
            ConfigOption(name="tags", default_value="", type_cast=str),
            ConfigOption(
                name="rules.disabled_rules",
                default_value=[],
                type_cast=utils.parse_disabled_rules,
            ),
            ConfigOption(
                name="stacktraces",
                default_value="ALL",
                type_cast=utils.parse_stacktraces_options,
            ),
        ]


class Inventory(ConfigBuilder):
    TOP_LEVEL = "inventory"

    def __init__(self):
        super(Inventory, self).__init__()

        self.default_options = [
            ConfigOption(
                name="analyze_libraries",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                name="enable", default_value=True, type_cast=utils.str_to_bool
            ),
            ConfigOption(name="tags", default_value="", type_cast=str),
        ]


class Protect(ConfigBuilder):
    TOP_LEVEL = "protect"

    def __init__(self):
        super(Protect, self).__init__()

        self.default_options = [
            ConfigOption(
                name="enable", default_value=True, type_cast=utils.str_to_bool
            ),
            ConfigOption(name="samples.probed", default_value=50, type_cast=int),
            ConfigOption(name="samples.blocked", default_value=25, type_cast=int),
            ConfigOption(name="samples.exploited", default_value=100, type_cast=int),
            ConfigOption(
                name="samples.blocked_at_perimeter", default_value=25, type_cast=int
            ),
            ConfigOption(
                name="rules.bot-blocker.enable",
                default_value=False,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                name="rules.cmd-injection.detect_chained_commands",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                name="rules.cmd-injection.detect_parameter_command_backdoors",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                name="rules.cmd-injection.mode",
                default_value=None,
                type_cast=utils.protect_mode_to_protobuf,
            ),
            ConfigOption(
                name="rules.disabled_rules",
                default_value=[],
                type_cast=utils.parse_disabled_rules,
            ),
            ConfigOption(
                name="rules.method-tampering.mode",
                default_value=None,
                type_cast=utils.protect_mode_to_protobuf,
            ),
            ConfigOption(
                name="rules.nosql-injection.mode",
                default_value=None,
                type_cast=utils.protect_mode_to_protobuf,
            ),
            ConfigOption(
                name="rules.path-traversal.detect_common_file_exploits",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                name="rules.path-traversal.detect_custom_code_accessing_system_files",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                name="rules.path-traversal.mode",
                default_value=None,
                type_cast=utils.protect_mode_to_protobuf,
            ),
            ConfigOption(
                name="rules.reflected-xss.mode",
                default_value=None,
                type_cast=utils.protect_mode_to_protobuf,
            ),
            ConfigOption(
                name="rules.sql-injection.detect_chained_queries",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                name="rules.sql-injection.detect_dangerous_functions",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                name="rules.sql-injection.detect_suspicious_unions",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                name="rules.sql-injection.detect_tautologies",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                name="rules.sql-injection.mode",
                default_value=None,
                type_cast=utils.protect_mode_to_protobuf,
            ),
            ConfigOption(
                name="rules.ssrf.mode",
                default_value=None,
                type_cast=utils.protect_mode_to_protobuf,
            ),
            ConfigOption(
                name="rules.unsafe-file-upload.mode",
                default_value=None,
                type_cast=utils.protect_mode_to_protobuf,
            ),
            ConfigOption(
                name="rules.xxe.mode",
                default_value=None,
                type_cast=utils.protect_mode_to_protobuf,
            ),
        ]


class Server(ConfigBuilder):
    TOP_LEVEL = "server"

    def __init__(self):
        super(Server, self).__init__()

        self.default_options = [
            ConfigOption(
                name="name", default_value=utils.get_hostname(), type_cast=str
            ),
            ConfigOption(name="path", default_value="/", type_cast=str),
            ConfigOption(name="type", default_value="", type_cast=str),
            ConfigOption(name="version", default_value="", type_cast=str),
            ConfigOption(name="environment", default_value="", type_cast=str),
            ConfigOption(name="tags", default_value="", type_cast=str),
        ]


class Root(ConfigBuilder):
    TOP_LEVEL = None

    def __init__(self):
        super(Root, self).__init__()

        self.default_options = [
            ConfigOption(name="enable", default_value=True, type_cast=utils.str_to_bool)
        ]



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
class SecurityException(Exception):

    STATUS_CODE = 403
    REASON_PHRASE = "Forbidden"
    STATUS = "{} {}".format(STATUS_CODE, REASON_PHRASE)

    def __init__(self, rule, message=None):
        message = (
            message
            if message
            else "Rule {} threw a security exception".format(rule.name)
        )
        super(SecurityException, self).__init__(message)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re
from contrast.extern.six import integer_types

from contrast.agent.policy import constants
from contrast.agent.assess.assess_exceptions import ContrastAssessException
from contrast.agent.assess.policy.trigger_actions.default_action import DefaultAction
from contrast.agent.assess.policy.trigger_actions import redos_action
from contrast.agent.assess.policy.trigger_actions import ssrf_action
from contrast.agent.policy.policy_node import PolicyNode


class TriggerNode(PolicyNode):

    TRIGGER_ACTIONS = {
        constants.TRIGGER_ACTION_DEFAULT: DefaultAction(),
        constants.TRIGGER_ACTION_REDOS: redos_action.RedosAction(),
        constants.TRIGGER_ACTION_SSRF: ssrf_action.SsrfAction(),
    }

    def __init__(
        self,
        module,
        class_name,
        instance_method,
        method_name,
        source,
        dataflow=True,
        good_value=None,
        bad_value=None,
        action=None,
        policy_patch=True,
        rule=None,
        python2_only=False,
        python3_only=False,
        protect_mode=False,
    ):
        super(TriggerNode, self).__init__(
            module,
            class_name,
            instance_method,
            method_name,
            source,
            None,
            policy_patch=policy_patch,
            python2_only=python2_only,
            python3_only=python3_only,
        )

        self.dataflow = dataflow

        self.good_value = (
            re.compile(good_value, flags=re.IGNORECASE) if good_value else None
        )
        self.bad_value = (
            re.compile(bad_value, flags=re.IGNORECASE) if bad_value else None
        )
        self.action = action or constants.TRIGGER_ACTION_DEFAULT

        self.rule = rule

        self.protect_mode = protect_mode

        self.validate()

    @property
    def node_type(self):
        return "TYPE_METHOD"

    @property
    def dataflow_rule(self):
        return self.dataflow

    def validate(self):
        super(TriggerNode, self).validate()

        if not self.dataflow_rule:
            return

        if not (self.sources and len(self.sources) != 0):
            raise ContrastAssessException(
                "Trigger {} did not have a proper source. Unable to create.".format(
                    self.name
                )
            )

    def get_matching_sources(self, self_obj, ret, args, kwargs):
        sources = []

        for source in self.sources:
            if source == constants.OBJECT:
                sources.append(self_obj)
            elif source == constants.ALL_ARGS:
                sources.append(args)
            elif source == constants.ALL_KWARGS:
                sources.append(kwargs)
            elif source == constants.RETURN:
                sources.append(ret)
            elif isinstance(source, integer_types) and len(args) > source:
                sources.append(args[source])
            elif kwargs and source in kwargs:
                sources.append(kwargs[source])

        return sources

    def get_protect_sources(self, args, kwargs):
        self_obj = args[0] if self.instance_method else None
        args = args[1:] if self.instance_method else args
        return self.get_matching_sources(self_obj, None, args, kwargs)

    @property
    def trigger_action(self):
        return self.TRIGGER_ACTIONS.get(self.action)

    @staticmethod
    def from_dict(obj, dataflow=True, rule=None):
        return TriggerNode(
            obj[constants.JSON_MODULE],
            obj.get(constants.JSON_CLASS_NAME, ""),
            obj.get(constants.JSON_INSTANCE_METHOD, True),
            obj[constants.JSON_METHOD_NAME],
            obj.get(constants.JSON_SOURCE, None),
            dataflow,
            obj.get(constants.JSON_GOOD_VALUE, None),
            obj.get(constants.JSON_BAD_VALUE, None),
            obj.get(constants.JSON_ACTION, None),
            policy_patch=obj.get(constants.JSON_POLICY_PATCH, True),
            rule=rule,
            python2_only=obj.get(constants.JSON_PY2_ONLY, False),
            python3_only=obj.get(constants.JSON_PY3_ONLY, False),
            protect_mode=obj.get(constants.JSON_PROTECT_MODE, False),
        )



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
import sys

from contrast.agent.assess.rules.config import (
    DjangoHttpOnlyRule,
    DjangoSecureFlagRule,
    DjangoSessionAgeRule,
)

from contrast.wsgi.middleware import WSGIMiddleware
from contrast.agent.middlewares.route_coverage.django_routes import (
    create_django_routes,
    build_django_route,
)
from contrast.utils.decorators import fail_safely

try:
    from django.urls import get_resolver
except ImportError:
    from django.core.urlresolvers import get_resolver


from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class DjangoWSGIMiddleware(WSGIMiddleware):
    """
    A subclass of the WSGI middleware that provides django route coverage and config
    scanning.

    This is not a Django-style middleware - it must wrap django's WSGI_APPLICATION,
    and does not belong in MIDDLEWARE / MIDDLEWARE_CLASSES.
    """

    def __init__(self, wsgi_app):
        self.app_name = self.get_app_name()

        self.config_rules = [
            DjangoSessionAgeRule(),
            DjangoSecureFlagRule(),
            DjangoHttpOnlyRule(),
        ]

        super(DjangoWSGIMiddleware, self).__init__(wsgi_app, self.app_name)

    def get_app_name(self):
        try:
            from django.conf import settings

            wsgi_application = settings.WSGI_APPLICATION

            return wsgi_application.split(".")[0]
        except Exception:
            return "Django Application"

    @fail_safely("Unable to get Django view func")
    def get_view_func(self, request):
        match = get_resolver().resolve(request.path_info or "/")
        if not match:
            return None

        return match.func

    @fail_safely("Unable to build route", return_value="")
    def build_route(self, view_func, url):
        return build_django_route(view_func)

    @fail_safely("Unable to get route coverage", return_value={})
    def get_route_coverage(self):
        """
        Route Coverage is the Assess feature that looks for routes generally defined
        in Django apps in a file like urls.py
        """
        return create_django_routes()

    @fail_safely("Failed to run config scanning rules")
    def scan_configs(self):
        """
        Run config scanning rules for assess

        Overridden from base class; gets called from base class
        """
        from django.conf import settings as app_settings

        app_config_module_name = os.environ.get("DJANGO_SETTINGS_MODULE")
        if not app_config_module_name:
            logger.warning("Unable to find Django settings for config scanning")
            return

        app_config_module = sys.modules.get(app_config_module_name)
        if not app_config_module:
            logger.warning("Django settings module not loaded; can't scan config")
            return

        for rule in self.config_rules:
            rule.apply(app_settings, app_config_module)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast.extern.six import iteritems
from contrast.api.dtm_pb2 import Activity

TEST_MODULES = (
    # deleting sample_namespace_lib_analysis from sys.modules will not allow us
    # to re import it in PY2. sample_namespace_lib_analysis shows up in sys.modules
    # as a builtin module. We only use it in one test, so its not necessary to do so anyway
    "sample_package_onefile",
    "sample_package_tld",
    "sample_package_tld.sample_tld",
    "sample_package_tld2",
    "sample_package_tld2.sample_tld2",
    "sample_package_tld2.subpackage",
    "sample_package_tld2.subpackage.import_file_in_subpackage",
    "sample_package_relative_imports",
    "sample_package_relative_imports.relative_import_same_dir_level",
    "sample_package_relative_imports.module_to_be_imported_in_samedir",
    "sample_package_relative_imports.subpackage",
    "sample_package_relative_imports.subpackage.sample_module_relative_import_parent_module",
    "sample_package_relative_imports.module_to_be_imported_by_subpackage_module",
)

TEST_MODULES_RELATIVE_IMPORT_DIST = (
    "sample_package_relative_imports/__init__.py",
    "sample_package_relative_imports/relative_import_same_dir_level.py",
    "sample_package_relative_imports/module_to_be_imported_in_samedir.py",
    "sample_package_relative_imports/subpackage/__init__.py",
    "sample_package_relative_imports/subpackage/sample_module_relative_import_parent_module.py",
    "sample_package_relative_imports/module_to_be_imported_by_subpackage_module.py",
)

TEST_MODULES_NAMESPACE_DIST = (
    "sample_namespace_lib_analysis/pkg2/sample_module_pkg2.py",
    "sample_namespace_lib_analysis/pkg2/__init__.py",
    "sample_namespace_lib_analysis/pkg1/sample_module_pkg1.py",
    "sample_namespace_lib_analysis/pkg1/__init__.py",
)

TEST_MODULES_MULTIPLE_TLDS_DIST = (
    "sample_package_tld/__init__.py",
    "sample_package_tld/sample_tld.py",
    "sample_package_tld2/__init__.py",
    "sample_package_tld2/sample_tld2.py",
    "sample_package_tld2/subpackage/__init__.py",
    "sample_package_tld2/subpackage/import_file_in_subpackage.py",
)

TEST_MODULE_ONEFILE = ("sample_package_onefile.py",)


def remove_sys_module_entries():
    for mod in TEST_MODULES:
        try:
            del sys.modules[mod]
        except:
            pass


def assert_files_loaded(activity, expected_modules_loaded, hash_filter=None):
    assert isinstance(activity, Activity)

    loaded_file_cnt = 0
    loaded_files = activity.library_usages

    for dist_hash, lib_update in iteritems(loaded_files):
        # leave hash_filter empty if you expect exactly expected_modules_loaded to be in activity.library_usages
        if hash_filter and dist_hash not in hash_filter:
            continue

        for file_loaded in lib_update.class_names.keys():
            assert file_loaded in expected_modules_loaded
            loaded_file_cnt += 1

    assert loaded_file_cnt == len(expected_modules_loaded)


def import_with_relative_imports():
    from sample_package_relative_imports import relative_import_same_dir_level
    from sample_package_relative_imports.subpackage import (
        sample_module_relative_import_parent_module,
    )

    relative_import_same_dir_level.SampleClass()
    sample_module_relative_import_parent_module.SampleClass()


def import_namespace_package():
    from sample_namespace_lib_analysis.pkg1 import sample_module_pkg1
    from sample_namespace_lib_analysis.pkg2 import sample_module_pkg2

    sample_module_pkg1.func_sample_namespace_pkg1()
    sample_module_pkg2.func_sample_namespace_pkg2()


def import_multiple_tlds():
    import sample_package_tld.sample_tld
    import sample_package_tld2.sample_tld2
    from sample_package_tld2.subpackage.import_file_in_subpackage import (
        SAMPLE_GLOBAL_VARIABLE_SUBPACKAGE,
    )

    sample_package_tld.sample_tld.SampleClassTLD()
    sample_package_tld2.sample_tld2.SampleClassTLD()
    x = SAMPLE_GLOBAL_VARIABLE_SUBPACKAGE  # pylint: disable=unused-variable


def import_sample_package_onefile():
    import sample_package_onefile

    sample_package_onefile.SampleClassOneFile()



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.agent import scope
from contrast.agent.policy.trigger_node import TriggerNode
from contrast.agent.middlewares.base_django_middleware import DjangoBaseMiddleware
from contrast.agent.middlewares.response_wrappers.django_response_wrapper import (
    DjangoResponseWrapper,
)
from contrast.agent.request_context import RequestContext
from contrast.utils.decorators import cached_property
from contrast.utils.exceptions.contrast_service_exception import (
    ContrastServiceException,
)

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class DjangoMiddleware(DjangoBaseMiddleware):
    """
    Legacy style Django Middleware. For Django versions 1.9 and below.

    It does not have the same __call__ style as previously. The old style has a method for the incoming request and
    outgoing response.
    """

    setting_name = "MIDDLEWARE_CLASSES"

    def __init__(self, *args):
        """
        Older style of Django Middleware; This middleware is for Django 1.9 and below
        (Django 1.10 is also supported but use DjangoMiddleware when possible instead)

        Unlike most framework middlewares, Django 1.9 versions and below call the __init__
        method of all middlewares on the first request made to the app, not when the app
        web server is initialized. This means that this middleware and all of its settings,
        including connection to Speedracer, will not happen until a first request is made to the app.

        Keeping the unused argument *args will allow this middleware to load even if used with the wrong
        Django middleware style. This way we can fail more gracefully, with a helpful message to the user.
        """
        self.validate_middleware()
        super(DjangoMiddleware, self).__init__()

        self.context = None

    def process_request(self, request):
        """
        Process request will only run the prefilter on the request context

        No work on the response is done here
        """
        if self.is_agent_enabled():
            body = self.get_body(request)
            if body is not None:
                context = RequestContext(request.environ, body)
                self.context = context

                # cannot use with contrast.CS__CONTEXT_TRACKER.lifespan(context) because we don't want
                # to delete the context until calling process_response
                contrast.CS__CONTEXT_TRACKER.set_current(context)
                return self.call_with_agent(context, request)

        return self.call_without_agent(request)

    def call_without_agent(self, request):
        """
        Normal without middleware call
        """
        super(DjangoMiddleware, self).call_without_agent()

    def call_with_agent(self, context, request):
        super(DjangoMiddleware, self).call_with_agent(context, request)

        try:
            self.prefilter(context)
        except ContrastServiceException as e:
            logger.warning(e)
            self.call_without_agent(request)
        except Exception as e:
            if self.settings.is_assess_enabled():
                contrast.STRING_TRACKER.ageoff()
            return self.handle_exception(e)

        return None

    def process_exception(self, request, exception):
        """
        Process exception is called when a view raises an exception

        We check if the exception is our own and handle it accordingly. We make sure to
        send any findings that may have occurred prior to the exception. This enables
        us to trigger findings even when the trigger function itself calls an
        exception.
        """
        try:
            return self.handle_exception(exception)
        except ContrastServiceException as e:
            logger.warning(e)
            return self.call_without_agent(request)
        except Exception:
            self.handle_ensure(self.context, request)
            raise

    def process_response(self, request, response):
        """
        Process response wraps the response and then runs the postfilter rules
        """
        if self.context is None:
            return response

        try:
            # This is to make sure that if an application catches our
            # SecurityException, we still block if necessary. Doing this before
            # trying to wrap the response makes sure that we handle the case
            # where the response is None because a SecurityException was raised
            # during process_view.
            self.check_for_blocked(self.context)

            with scope.contrast_scope():
                cs__response = DjangoResponseWrapper(response)

            self.context.extract_response_to_context(cs__response)

            self.postfilter(self.context)

            self.context.timer.set_end("total")

            return response

        except Exception as e:
            return self.handle_exception(e)
        finally:
            self.handle_ensure(self.context, request)
            if request is not None:
                self.log_end_request_analysis(request.path)
            if self.settings.is_assess_enabled():
                contrast.STRING_TRACKER.ageoff()
            contrast.CS__CONTEXT_TRACKER.delete_current()
            self.context = None

    @cached_property
    def trigger_node(self):
        """
        Trigger node property used by assess reflected xss postfilter rule
        """
        try:
            from django.apps import apps

            app = apps.app_configs[self.app_name]
            module = app.module.__name__
            method_name = app.name
        except Exception:
            logger.warning("Failed to get django app metadata for xss trigger")
            module = "<app module>"
            method_name = "<django app>"

        return TriggerNode(module, "", False, method_name, "RETURN"), ()



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.utils.exceptions.security_exception import SecurityException

from contrast.extern import structlog as logging
import traceback

logger = logging.getLogger("contrast")


def set_context(flag):
    """
    Decorator to set the context of a method. For instance, multiple methods call the same method but it may want
    to handle some logic differently but the same method does not know who is calling it.

    As long as the attribute, represented as a string exists in the object it will flip it before and after the call.

    :param flag: string of attribute
    :return: decorator that takes in one argument
    """

    def wrap(func):
        def wrapped_func(self, *args, **kwargs):
            if hasattr(self, flag):
                setattr(self, flag, not getattr(self, flag))

            try:
                result = func(self, *args, **kwargs)
            finally:
                if hasattr(self, flag):
                    setattr(self, flag, not getattr(self, flag))

            return result

        return wrapped_func

    return wrap


def fail_safely(log_message, log_level="exception", return_value=None):
    """
    Decorator that will run the decorated function/method and, if
    an exception is raised, return a safe value and log the error.

    Note that SecurityException will always be re-raised.

    :param log_message: message to log in case of failure
    :param log_level: level to log in case of failure
    :param return_value: safe value to return in case of failure
    :return: original func return or return_value
    """

    def wrap(original_func):
        def run_safely(*args, **kwargs):
            try:
                return original_func(*args, **kwargs)
            except SecurityException:
                raise
            except Exception as e:
                try:
                    getattr(logger, log_level)(log_message + ": " + str(e))
                    logger.debug("wrapped function args: %s", args)
                    logger.debug("wrapped function kwargs: %s", kwargs)
                except Exception as e:
                    logger.debug(
                        "Failed to log arguments passed to original_func",
                        exception_thrown=str(e),
                    )

            return return_value

        return run_safely

    return wrap


def fail_quietly(log_message, return_value=None):
    """
    Similar to fail_safely (see above)

    This decorator is intended to handle cases where an exception may occur but won't
    disrupt normal operation of the agent. This decorator should be used to protect
    against external exceptions we can't prevent but still want to handle.

    In these cases, we log an error message and the exception traceback, both at DEBUG
    level.
    """

    def wrap(original_func):
        def run_safely(*args, **kwargs):
            try:
                return original_func(*args, **kwargs)
            except SecurityException:
                raise
            except Exception as e:
                logger.debug("%s: %s", log_message, str(e))
                logger.debug(traceback.format_exc())

            return return_value

        return run_safely

    return wrap


class cached_property(object):
    """
    https://github.com/pydanny/cached-property
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self

        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Database adapter patch for psycopg2.
This module's cursor doesn't have an `executescript` method.
"""

import os
from contrast.extern.wrapt import register_post_import_hook

import contrast
from contrast.patches.databases import dbapi2
from contrast.utils import inventory_utils
from contrast.utils.patch_utils import patch_cls_or_instance

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")

PSYCOPG2 = "psycopg2"
VENDOR = "PostgreSQL"
PSYCOPG2_EXTENSIONS = "psycopg2.extensions"
CONNECT = "connect"


def patch_connect(module):
    """
    Record DB inventory for a Postgres connection.

    Here we make a good effort to find connection params. There are several ways that these can be set,
    in the following order of priority (using dbname as an example):
    - using the `connection_factory` kwarg
    - as a kwarg itself - `dbname` or the deprecated `database`
    - via the dbname parameter in the dsn string
    - with the PGDATABASE environment variable

    Newer versions of psycopg2 (v2.7, ~2017) support connection.get_dsn_parameters, which provides
    a dictionary of the parsed connection params - we're interested in `dbname`.

    For now, it's still possible for us to miss the dbname (i.e. an old version of psycopg2 using
    the dsn string only), but this is unlikely and it would only affect inventory.
    """

    def _connect(orig_func, patch_policy=None, *args, **kwargs):
        connection = orig_func(*args, **kwargs)
        try:
            context = contrast.CS__CONTEXT_TRACKER.current()
            if context is not None:
                host, port, dbname = _get_inventory_params(connection, kwargs)
                db_inventory = dict(
                    vendor=VENDOR, host=host, port=port, database=dbname
                )
                inventory_utils.append_db(context.activity, db_inventory)
        except Exception:
            logger.exception("Failed to add inventory for %s", VENDOR)

        return connection

    patch_cls_or_instance(module, "connect", _connect)


def _get_inventory_params(connection, kwargs):
    """
    Get the host, port, and dbname for the current connection. This is used for inventory reporting.
    """
    dsn_params = getattr(connection, "get_dsn_parameters", lambda: {})()
    host = (
        dsn_params.get("host")
        or kwargs.get("host")
        or os.environ.get("PGHOST", "unknown_host")
    )
    port = (
        dsn_params.get("port")
        or kwargs.get("port")
        or os.environ.get("PGPORT", "unknown_port")
    )
    dbname = (
        dsn_params.get("dbname")
        or kwargs.get("dbname")
        or kwargs.get("database")
        or os.environ.get("PGDATABASE", "unknown_database")
    )
    return host, port, dbname


def instrument_psycopg2(psycopg2):
    patch_connect(psycopg2)


def instrument_extensions(psycopg2_extensions):
    """
    psycopg2 allows for custom cursor objects. We can still handle these as long as they
    subclass the base/default cursor provided by psycopg2 and use super() to call the original
    execute* methods. This seems to be the only possible use case, so we'll always have
    execute instrumented.

    see https://www.psycopg.org/docs/advanced.html#subclassing-connection

    Also note that psycopg2 includes several convenient extensions to dbapi2:
    - psycopg2.extras.execute_batch
    - psycopg2.extras.execute_values
    These both call `execute` internally
    """
    dbapi2.instrument_cursor(PSYCOPG2, psycopg2_extensions.cursor)


def register_patches():
    register_post_import_hook(instrument_psycopg2, PSYCOPG2)
    register_post_import_hook(instrument_extensions, PSYCOPG2_EXTENSIONS)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


from . import (
    cs_cstringio,
    cs_io,
    re_patch,
    threading_patch,
    exec_and_eval,
    lxml_patch,
    pathlib_patch,
    pycryptodome_patch,
    pyramid_patch,
    sys_patch,
    library_reader_import_patches,
)

from .databases import (
    mysql_connector_patch,
    pymysql_patch,
    psycopg2_patch,
    sqlalchemy_patch,
    sqlite3_patch,
)

from .frameworks import bottle_patches, django_patches, drf_patches


COMMON_PATCH_MODULES = (
    # pycrypto_patch,
    pathlib_patch,
    pycryptodome_patch,
    # pycryptodomex_patch,
    sqlalchemy_patch,
    # our sqlite3_patch also contains the import hook for pysqlite2.dbapi2
    sqlite3_patch,
    mysql_connector_patch,
    pymysql_patch,
    psycopg2_patch,
)


LIBRARY_READER_PATCHES = (library_reader_import_patches,)


ASSESS_PATCH_MODULES = (
    cs_cstringio,
    cs_io,
    re_patch,
    exec_and_eval,
    lxml_patch,
    pyramid_patch,
    django_patches,
    drf_patches,
    bottle_patches,
    threading_patch,
    sys_patch,
)


def register_module_patches(module, patch_group):
    logger.debug("registering %s patches for %s", patch_group, module.__name__)

    try:
        module.register_patches()
    except Exception:
        logger.exception("failed to register patches for %s", module.__name__)


def register_library_patches():
    for module in LIBRARY_READER_PATCHES:
        register_module_patches(module, "library analysis")


def register_common_patches():
    for module in COMMON_PATCH_MODULES:
        register_module_patches(module, "common")


def register_assess_patches():
    for module in ASSESS_PATCH_MODULES:
        register_module_patches(module, "assess")



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern.six import iteritems

import contrast
from contrast.agent.assess.apply_trigger import cs__apply_trigger


def apply(
    rule, trigger_nodes, ret, args, kwargs=None
):  # pylint: disable=redefined-builtin
    """
    Iterates over all given trigger policy nodes and applies the given rule

    This method gets called from within a patched trigger function in order
    to determine whether the given rule has been violated.

    @param rule: `TriggerRule` instance representing rule to be evaluated
    @param trigger_nodes: List of `TriggerNode` instances
    @param ret: Result returned by the trigger function
    @param args: Tuple containing args passed to trigger function
    @param kwargs: Dict containing kwargs passed to trigger function
    """
    if kwargs is None:
        kwargs = {}

    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None:
        return

    for node in trigger_nodes:
        if node.instance_method:
            self_obj = args[0]
            args = args[1:]  # args[0] is `self` for instance methods
        else:
            self_obj = None  # module-level functions do not have a self
            args = args

        possible_sources = node.get_matching_sources(self_obj, ret, args, kwargs)

        for source in possible_sources:
            if isinstance(source, dict):
                for key, value in iteritems(source):
                    cs__apply_trigger(
                        context, rule, node, key, self_obj, ret, None, args, kwargs
                    )
                    # pass in the key here for building_finding
                    cs__apply_trigger(
                        context, rule, node, value, self_obj, ret, key, args, kwargs
                    )
            elif isinstance(source, (tuple, list)):
                for item in source:
                    cs__apply_trigger(
                        context, rule, node, item, self_obj, ret, None, args, kwargs
                    )
            else:
                cs__apply_trigger(
                    context, rule, node, source, self_obj, ret, None, args, kwargs
                )
        else:
            if not rule.dataflow:
                cs__apply_trigger(
                    context, rule, node, None, self_obj, ret, None, args, kwargs
                )



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.middlewares.route_coverage.coverage_utils import (
    CoverageUtils,
    DEFAULT_ROUTE_METHODS,
)


def create_flask_routes(app):
    """
    Returns all the routes registered to a Flask app as a dict
    :param app: Flask app
    :return: dict {route_id:  RouteCoverage}
    """
    routes = {}

    with app.app_context():
        for rule in list(app.url_map.iter_rules()):
            view_func = app.view_functions[rule.endpoint]

            route = build_flask_route(rule.endpoint, view_func)

            route_id = str(id(view_func))

            methods = rule.methods or DEFAULT_ROUTE_METHODS

            for method_type in methods:
                key = CoverageUtils.build_key(route_id, method_type)
                routes[key] = CoverageUtils.build_route_coverage(
                    verb=method_type,
                    url=CoverageUtils.get_normalized_uri(str(rule)),
                    route=route,
                )

    return routes


def build_flask_route(view_func_name, view_func):
    view_func_args = CoverageUtils.build_args_from_function(view_func)
    return view_func_name + view_func_args



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.request import Request
from contrast.test import helper


class ExampleRequest(Request):
    def __init__(self, environ=None):
        super(Request, self).__init__(environ or helper.get_simple_request())



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.applies import apply_rule
from contrast.utils.patch_utils import patch_cls_or_instance

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


def protect_patch(original_func, patch_policy, *args, **kwargs):
    """
    Protect patch that will run in addition to running original_func.
    If we cannot run the protect rule, at the very least run the original_func.
    """
    return apply_rule(patch_policy, original_func, args, kwargs)


def apply_protect_patch(patch_site, patch_policy):
    logger.debug("Applying protect patch to %s", patch_policy.name)

    patch_cls_or_instance(
        patch_site, patch_policy.method_name, protect_patch, patch_policy=patch_policy
    )



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
class ObjectShare(object):
    # Strings
    AND = "&"
    PIPE = "|"
    ASTERISK = "*"
    BACK_SLASH = "\\"
    EMPTY_STRING = ""
    COLON = ":"
    COMMA = ","
    DASH = "-"
    DOUBLE_QUOTE = '"'
    EQUALS = "="
    EXCLAMATION = "!"
    HTTP_SCORE = "HTTP_"
    HTTP_START = "http:"
    HTTPS_START = "https:"
    NEW_LINE = "\n"
    NONE_STRING = "None"
    PERIOD = "."
    PLUS_SIGN = "+"
    POUND_SIGN = "#"
    QUESTION_MARK = "?"
    RETURN = "\r"
    SEMICOLON = ";"
    SINGLE_QUOTE = "'"
    SLASH = "/"
    SPACE = " "
    UNDERSCORE = "_"
    AT = "@"
    LEFT_ANGLE = "<"
    RIGHT_ANGLE = ">"
    LEFT_BRACKET = "["
    RIGHT_BRACKET = "]"
    LEFT_PAREN = "("
    RIGHT_PAREN = ")"
    BACK_TICK = "`"

    DOUBLE_QUOTE_ESCAPED = "&quot;"

    CLASS = "class"
    CONTRAST_ASSESS_METHOD_START = "cs__assess_"
    CONTRAST_METHOD_START = "cs__"

    MODULE = "module"

    READ_FLAG = "r"
    WRITE_FLAG = "w"
    READ_WRITE_FLAG = "rw"

    PARENT_PATH = ".."

    CACHE = "cache"

    EMPTY_ARRAY = []

    UTF_8_ENCODING = "utf-8"

    # Messages
    OVERRIDE_MESSAGE = (
        "A security filter prevented original response from being returned."
    )



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern.six import iteritems
from contrast.api.dtm_pb2 import Activity, Finding as FindingDtm

from contrast.utils.string_utils import protobuf_safe

from contrast.agent.policy.constants import (
    CURRENT_FINDING_VERSION,
    MINIMUM_FINDING_VERSION,
)


from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class Finding(object):
    """
    Wrapper around api.dtm_pb2.Finding
    """

    def __init__(self, rule, properties, **kwargs):
        self.finding = FindingDtm()

        rule.add_events_to_finding(self.finding, **kwargs)

        self.finding.rule_id = protobuf_safe(rule.name)

        for key, value in iteritems(properties):
            self.finding.properties[key] = value

        hash_code = rule.generate_preflight_hash(**kwargs)

        self.finding.hash_code = hash_code
        self.finding.preflight = ",".join([rule.name, hash_code])

        logger.debug("Created finding for %s", rule.name)
        logger.debug("initial preflight value: %s", self.finding.preflight)

    def get_finding_msg(self):
        return self.finding

    @staticmethod
    def pick_version(finding):
        """
        Given a finding dtm message, determine the finding version.

        Note:  This is a staticmethod and not applied when the create method is called
        because sometimes the finding version cannot be determined until later in the
        request lifecycle.

        :param finding: instance of dtm_pb2.Finding
        :return: finding version
        """
        # dataflow or non-dataflow finding with route(s)
        if finding.routes:
            return CURRENT_FINDING_VERSION

        # non-dataflow rules without routes
        if not finding.events:
            return CURRENT_FINDING_VERSION

        # dataflow finding without routes
        return MINIMUM_FINDING_VERSION


def send_finding(finding, context=None):
    """
    Send a finding by either appending it to request context
    OR sending it immediately.

    :param finding:  api.dtm_pb2.Finding instance
    :param context: Request context instance
    :return: None
    """
    # Import here to prevent circular import.
    from contrast.agent.service_client import send_messages
    
    if context:
        # Once RequestContext is deprecated, this flow can be removed.
        context.activity.findings.extend([finding])
    else:
        activity = Activity()
        activity.findings.extend([finding])

        send_messages([activity])



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.agent import scope
from contrast.agent.assess.policy.utils import build_method_name, logger
from contrast.agent.assess.policy.analysis import analyze
from contrast.extern.wrapt import ObjectProxy


def assess_method(original_method, patch_policy, *args, **kwargs):
    """
    Patching method to replace old method and call our assess code with the original method
    :param original_method: method to call for result
    :param patch_policy: PatchLocationPolicy containing all policy nodes for this patch
    :param args: method args
    :param kwargs: method kwargs
    :return: result of original method
    """
    context = contrast.CS__CONTEXT_TRACKER.current()

    try:
        result = original_method(*args, **kwargs)
    except Exception:
        result = None
        raise
    finally:
        analyze(context, patch_policy, result, args, kwargs)

    return result


def assess_classmethod(original_method, patch_policy, *args, **kwargs):
    """
    Patching method to replace old method and call our assess code with the original method
    :param original_method: method to call for result
    :param patch_policy: PatchLocationPolicy containing all policy nodes for this patch
    :param args: method args
    :param kwargs: method kwargs
    :return: result of original method

    A separate method was required for classmethod patch because we need to remove
    argument 1. arg 1 is the class. This is something that is automatically passed to
    the function so passing it again will cause a TypeError.
    """
    context = contrast.CS__CONTEXT_TRACKER.current()

    try:
        result = original_method(*args[1:], **kwargs)
    except Exception:
        result = None
        raise
    finally:
        analyze(context, patch_policy, result, args, kwargs)

    return result


def assess_deadzone(original_method, patch_policy, *args, **kwargs):
    """
    Patching method to replace old method and call the old method in contrast scope,
    preventing any analysis down the stack.

    :param original_method: method to call for result
    :param patch_policy: PatchLocationPolicy containing all policy nodes for this patch
    :param args: method args
    :param kwargs: method kwargs
    :return: result of original method
    """
    with scope.contrast_scope():
        return original_method(*args, **kwargs)


def assess_property(original_property_name, patch_policy, *args, **kwargs):
    """
    Calls the original property by looking for it in in the cs_assess_{property} location
    to return the property value, and run assess analysis.
    """
    context = contrast.CS__CONTEXT_TRACKER.current()
    try:
        cls_instance = args[0]
        cs_method_name = build_method_name(original_property_name)
        result = getattr(cls_instance, cs_method_name)
    except Exception:
        result = None
        raise
    finally:
        analyze(context, patch_policy, result, args, kwargs)

    return result


def apply_cached_property(cls_or_module, patch_policy, property_name, orig_property):
    """
    Older werkzeug versions implement cached_property that does not inherit from property.
    This causes us to have to use a workaround for patching to avoid errors.

    Instead of replacing the cached_property with a new property, we replace it with
    and object proxy with a custom __get__ method.
    """
    proxied_property = CachedPropertyProxy(orig_property, property_name, patch_policy)

    try:
        setattr(cls_or_module, property_name, proxied_property)
    except Exception:
        logger.exception("Failed to apply patch to cached_property: %s", property_name)

    return True


class CachedPropertyProxy(ObjectProxy):
    cs__attr_name = None
    cs__patch_policy = None

    def __init__(self, wrapped, attr_name, patch_policy):
        super(CachedPropertyProxy, self).__init__(wrapped)
        self.cs__patch_policy = patch_policy
        self.cs__attr_name = attr_name

    def __get__(self, *args, **kwargs):
        context = contrast.CS__CONTEXT_TRACKER.current()
        result = self.__wrapped__.__get__(*args, **kwargs)

        try:
            # Self is the only arg that seems to be relevant for policy/reporting
            args = (self.__wrapped__,)
            analyze(context, self.cs__patch_policy, result, args, {})
        except Exception:
            logger.exception("Failed to apply policy for %s", self.cs__attr_name)

        return result



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.agent.policy.trigger_node import TriggerNode
from contrast.agent.middlewares.base_middleware import BaseMiddleware
from contrast.agent.middlewares.environ_tracker import track_environ_sources
from contrast.agent.middlewares.route_coverage.pyramid_routes import PyramidRoutesMixin

from contrast.agent.request_context import RequestContext
from contrast.utils.decorators import cached_property
from contrast.utils.exceptions.contrast_service_exception import (
    ContrastServiceException,
)

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")

WEBOB = "webob"


class PyramidMiddleware(PyramidRoutesMixin, BaseMiddleware):
    def __init__(self, handler, registry):
        self.registry = registry
        self.app_name = self.get_app_name()

        super(PyramidMiddleware, self).__init__()

        self.handler = handler

    def get_app_name(self):
        try:
            return self.registry.package_name
        except Exception:
            return "pyramid_app"

    def __call__(self, request):
        if self.is_agent_enabled() and request:
            context = RequestContext(request.environ)

            if self.settings.is_assess_enabled():
                track_environ_sources(WEBOB, context, request.environ)

            with contrast.CS__CONTEXT_TRACKER.lifespan(context):
                return self.call_with_agent(request, context)

        return self.call_without_agent(request)

    def call_without_agent(self, request):
        """
        Normal without middleware call
        """
        super(PyramidMiddleware, self).call_without_agent()
        return self.handler(request)

    def call_with_agent(self, request, context):
        self.log_start_request_analysis(request.environ.get("PATH_INFO"))

        context.timer.set_start("total")

        try:
            self.prefilter(context)

            logger.debug("Start app code and get response")
            response = self.handler(request)
            logger.debug("Finished app code and get response")

            # Pyramid's response class is based on Webob's, which already
            # implements BaseResponseWrapper's requirements
            context.extract_response_to_context(response)

            self.postfilter(context)

            self.check_for_blocked(context)

            context.timer.set_end("total")

            return response
        except ContrastServiceException as e:
            logger.warning(e)
            return self.call_without_agent(request)
        except Exception as e:
            return self.handle_exception(e)
        finally:
            self.handle_ensure(context, request)
            self.log_end_request_analysis(context.request.path)
            if self.settings.is_assess_enabled():
                contrast.STRING_TRACKER.ageoff()

    @cached_property
    def trigger_node(self):
        method_name = self.handler.__name__

        module, class_name, args, instance_method = self._process_trigger_handler(
            self.handler
        )

        return (
            TriggerNode(module, class_name, instance_method, method_name, "RETURN"),
            args,
        )

    def generate_security_exception_response(self):
        from pyramid.httpexceptions import HTTPForbidden

        return HTTPForbidden(explanation=self.OVERRIDE_MESSAGE)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast.extern.six import PY2

from contrast.extern.wrapt import register_post_import_hook
from contrast.agent.policy import patch_manager
from contrast.utils.patch_utils import patch_cls_or_instance


def noop_intern(orig_func, patch_policy, string):
    """
    Defeat interning by returning the original string

    sys.intern takes no kwargs
    """
    return string


def patch_sys(module):
    if not PY2:
        # sys.intern is PY3 only
        patch_cls_or_instance(module, "intern", noop_intern)


def register_patches():
    register_post_import_hook(patch_sys, "sys")


def reverse_patches():
    # sys is always imported, no need to check
    patch_manager.reverse_patches_by_owner(sys)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
try:
    from django.urls import re_path as compat_url
except ImportError:
    from django.conf.urls import url as compat_url

from contrast.test.framework_test import django_views as views

framework_test_urlpatterns = [
    compat_url(r"^$", views.home),
    compat_url(r"^static-image/?$", views.static_image),
    compat_url(r"^xss-autoescape/?$", views.django_autoescape),
    compat_url(r"^xss-autoescape-safe/?$", views.django_autoescape_safe),
    compat_url(r"^xss-no-autoescape/?$", views.django_no_autoescape),
    compat_url(r"^file-contents/?$", views.file_contents),
    compat_url(r"^uce-base64/?$", views.uce_base64),
    compat_url(r"^cmdi-file/?$", views.cmdi_file),
    compat_url(r"^nosqli/?$", views.nosqli),
    compat_url(r"^stream-source/?$", views.stream_source),
    compat_url(r"^cookie-source/?$", views.cookie_source),
    compat_url(r"^header-source/?$", views.header_source),
    compat_url(r"^header-key-source/?$", views.header_key_source),
    compat_url(r"^xss/?$", views.xss),
    compat_url(r"^xss-jinja/?$", views.xss_jinja),
    compat_url(r"^xss-mako/?$", views.xss_mako),
    compat_url(r"^xss-cookie/?$", views.xss_cookie),
    compat_url(r"^xss-csv/?$", views.xss_csv),
    compat_url(r"^raw-xss-streaming/?$", views.raw_xss_streaming),
    compat_url(r"^logout/?$", views.logout),
    compat_url(r"^unvalidated-redirect/?$", views.unvalidated_redirect),
    compat_url(r"^two-vulns/?$", views.two_vulns),
    compat_url(r"^dynamic-sources/?$", views.dynamic_sources),
    compat_url(r"^stored-xss/?$", views.stored_xss),
    compat_url(r"^xpath-injection/?$", views.xpath_injection),
    compat_url(r"^trust-boundary-violation/?$", views.trust_boundary_violation),
    compat_url(r"^crypto-bad-ciphers/?$", views.crypto_bad_ciphers),
    compat_url(r"^streaming/?$", views.streaming),
    compat_url(r"^dynamic_url_view/?$", views.dynamic_url_view),
    compat_url(
        r"^import_package_with_relative_imports/?$",
        views.import_package_with_relative_imports,
    ),
    compat_url(r"^import_namespace_package/?$", views.import_sample_namespace_package),
    compat_url(r"^import_package_onefile/?$", views.import_package_onefile),
    compat_url(
        r"^import_sample_dist_multiple_tlds/?$", views.import_sample_dist_multiple_tlds
    ),
]



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.wsgi.middleware import WSGIMiddleware
from contrast.utils.decorators import fail_safely, fail_quietly
from contrast.extern import structlog as logging
from contrast.agent.middlewares.route_coverage.bottle_routes import (
    create_bottle_routes,
    build_bottle_route,
)

logger = logging.getLogger("contrast")


class BottleMiddleware(WSGIMiddleware):
    # Since Bottle is WSGI-based, there is no way to retrieve the app name.
    # Use common config to define an app name.
    def __init__(self, app):
        super(BottleMiddleware, self).__init__(app, app_name="Bottle Application")

    @fail_safely("Unable to get route coverage", return_value={})
    def get_route_coverage(self):
        return create_bottle_routes(self.wsgi_app)

    @fail_quietly("Unable to get Bottle view func")
    def get_view_func(self, request):
        path = request.path
        if not path:
            return None
        method = request.method
        route_info = self.wsgi_app.match({"PATH_INFO": path, "REQUEST_METHOD": method})
        if not route_info:
            return None
        view_func = route_info[0].callback
        return view_func

    @fail_safely("Unable to build route", return_value="")
    def build_route(self, view_func, url):
        return build_bottle_route(url, view_func)
